"""Base classes which define the ELPP Associations"""

import copy
import logging
import re
from collections import defaultdict
from os.path import basename, split, splitext

from stpipe.format_template import FormatTemplate

from romancal.associations import libpath
from romancal.associations.association import Association
from romancal.associations.exceptions import AssociationNotValidError
from romancal.associations.lib.acid import ACID
from romancal.associations.lib.constraint import Constraint, SimpleConstraint
from romancal.associations.lib.counter import Counter
from romancal.associations.lib.dms_base import (
    _EMPTY,
    IMAGE2_NONSCIENCE_EXP_TYPES,
    IMAGE2_SCIENCE_EXP_TYPES,
    SPEC2_SCIENCE_EXP_TYPES,
    WFI_DETECTORS,
    DMSAttrConstraint,
    DMSBaseMixin,
)
from romancal.associations.lib.keyvalue_registry import KeyValueRegistryError
from romancal.associations.lib.member import Member
from romancal.associations.lib.process_list import ProcessList
from romancal.associations.lib.product_utils import (
    prune_duplicate_associations,
    prune_duplicate_products,
)
from romancal.associations.lib.utilities import evaluate, is_iterable
from romancal.associations.registry import RegistryMarker

__all__ = [
    "ASN_SCHEMA",
    "AsnMixin_AuxData",
    "AsnMixin_Science",
    "AsnMixin_Spectrum",
    "AsnMixin_Lv2FOV",
    "AsnMixin_Lv2Image",
    "AsnMixin_Lv2GBTDSfull",
    "AsnMixin_Lv2GBTDSpass",
    "Constraint",
    "Constraint_Base",
    "Constraint_Category",
    "Constraint_Expos",
    "Constraint_Image",
    "Constraint_Instrument",
    "Constraint_Obsnum",
    "Constraint_Optical_Path",
    "Constraint_Pass",
    "Constraint_Spectral",
    "Constraint_SubCategory",
    "Constraint_Tile",
    "Constraint_Image_Science",
    "Constraint_Sequence",
    "Constraint_Single_Science",
    "Constraint_Spectral_Science",
    "Constraint_Target",
    "Constraint_Filename",
    "DMS_ELPP_Base",
    "DMSAttrConstraint",
    "ProcessList",
    "SimpleConstraint",
    "Utility",
]
# Configure logging
logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())

# The schema that these associations must adhere to.
ASN_SCHEMA = RegistryMarker.schema(libpath("asn_schema_jw_level3.json"))

# DMS file name templates
_LEVEL1B_REGEX = r"(?P<path>.+)(?P<type>_uncal)(?P<extension>\..+)"
_DMS_POOLNAME_REGEX = r"jw(\d{5})_(\d{8}[Tt]\d{6})_pool"

# Product name regex's
_REGEX_ACID_VALUE = r"(o\d{3}|(c|a)\d{4})"

# Exposures that should have received Level2b processing
LEVEL2B_EXPTYPES = []
LEVEL2B_EXPTYPES.extend(IMAGE2_SCIENCE_EXP_TYPES)
LEVEL2B_EXPTYPES.extend(IMAGE2_NONSCIENCE_EXP_TYPES)
LEVEL2B_EXPTYPES.extend(SPEC2_SCIENCE_EXP_TYPES)

# Association Candidates that should never make Level3 associations
INVALID_AC_TYPES = ["background"]


class DMS_ELPP_Base(DMSBaseMixin, Association):
    """Basic class for DMS Level associations."""

    # Set the validation schema
    schema_file = ASN_SCHEMA.schema

    # Attribute values that are indicate the
    # attribute is not specified.
    INVALID_VALUES = _EMPTY

    # Make sequences type-dependent
    _sequences = defaultdict(Counter)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # Initialize validity checks
        self.validity.update(
            {
                "has_science": {
                    "validated": True,
                    "check": lambda member: member["exptype"] == "science",
                },
            }
        )

        # Other presumptions on the association
        if "constraints" not in self.data:
            self.data["constraints"] = "No constraints"
        if "asn_type" not in self.data:
            self.data["asn_type"] = "user_built"
        if "asn_id" not in self.data:
            self.data["asn_id"] = "a3001"
        if "target" not in self.data:
            self.data["target"] = "none"
        if "asn_pool" not in self.data:
            self.data["asn_pool"] = "none"

    @property
    def current_product(self):
        return self.data["products"][-1]

    def members_by_type(self, member_type):
        """Get list of members by their exposure type"""
        member_type = member_type.lower()
        try:
            members = self.current_product["members"]
        except KeyError:
            result = []
        else:
            result = [
                member for member in members if member_type == member["exptype"].lower()
            ]

        return result

    def has_science(self):
        """Does association have a science member

        -------
        bool
            True if it does.
        """
        limit_reached = len(self.members_by_type("science")) >= 1
        return limit_reached

    def __eq__(self, other):
        """Compare equality of two associations"""
        if isinstance(other, DMS_ELPP_Base):
            result = self.data["asn_type"] == other.data["asn_type"]
            result = result and (self.member_ids == other.member_ids)
            return result

        return NotImplemented

    def __ne__(self, other):
        """Compare inequality of two associations"""
        result = self.__eq__(other)
        if result is not NotImplemented:
            result = not result
        return result

    @property
    def dms_product_name(self):
        """Define product name.

        Returns
        -------
        product_name : str
            The product name
        """
        return self._dms_product_name(self)

    @staticmethod
    def _dms_product_name(association):
        """Define product name.

        Parameters
        ----------
        association : `Association`
            Association to get the name from.

        Returns
        -------
        product_name : str
            The product name
        """
        target = association._get_target()

        instrument = association._get_instrument()

        opt_elem = association._get_opt_element()

        exposure = association._get_exposure()
        if exposure:
            exposure = "-" + exposure

        subarray = association._get_subarray()
        if subarray:
            subarray = "-" + subarray

        product_name = (
            "r{program}-{acid}" "_{target}" "_{instrument}" "_{opt_elem}{subarray}"
        )
        if "Full" in association.data["asn_rule"]:
            subarray = "Full"

        if "Pass" in association.data["asn_rule"]:
            subarray = "Pass"

        product_name = product_name.format(
            program=association.data["visit_id"],
            acid=association.acid.id,
            target=target,
            instrument=instrument,
            opt_elem=opt_elem,
            subarray=subarray,
            exposure=exposure,
        )

        return product_name.lower()

    def update_asn(self, item=None, member=None):
        """Update association meta information

        Parameters
        ----------
        item : dict or None
            Item to use as a source. If not given, item-specific
            information will be left unchanged.

        member : Member or None
            An association member to use as source.
            If not given, member-specific information will be update
            from current association/product membership.

        Notes
        -----
        If both `item` and `member` are given,
        information in `member` will take precedence.
        """
        super().update_asn(item=item, member=member)

        # Constraints
        self.data["constraints"] = str(self.constraints)

        # ID
        self.data["asn_id"] = self.acid.id

        # Target
        self.data["target"] = self._get_target()

        # Item-based information
        if item is not None:
            # Program
            if self.data["program"] == "noprogram":
                self.data["program"] = f"{item['program']:0>5s}"

            # Pool
            if self.data["asn_pool"] == "none":
                self.data["asn_pool"] = basename(item.meta["pool_file"])
                parsed_name = re.search(
                    _DMS_POOLNAME_REGEX, self.data["asn_pool"].split(".")[0]
                )
                if parsed_name is not None:
                    pool_meta = {
                        "program_id": parsed_name.group(1),
                        "version": parsed_name.group(2),
                    }
                    self.meta["pool_meta"] = pool_meta

        # Product-based updates
        product = self.current_product
        product["name"] = self.dms_product_name

    def make_member(self, item):
        """Create a member from the item

        Parameters
        ----------
        item : dict
            The item to create member from.

        Returns
        -------
        member : Member
            The member
        """
        try:
            exposerr = item["exposerr"]
        except KeyError:
            exposerr = None

        # Get exposure type
        exptype = self.get_exposure_type(item)

        # Determine expected member name
        expname = Utility.rename_to_level2(
            item["filename"], exp_type=item["exp_type"], member_exptype=exptype
        )

        member = Member(
            {
                "expname": expname,
                "exptype": exptype,
                "asn_candidate": item["asn_candidate"],
                "exposerr": exposerr,
            },
            item=item,
        )
        return member

    def make_fov_asn(self):
        """Take the association with an single exposure with _WFI_ in the name
              and expand that to include all 18 detectors.

        Returns
        -------
        associations : [association[, ...]]
            List of new members to be used in place of
            the current one.
        """
        results = []

        # expand the products from _wfi_ to _wfi{det}_
        for product in self["products"]:
            for member in product["members"]:
                asn = copy.deepcopy(self)
                asn.data["products"] = None
                product_name = (
                    splitext(
                        split(self.data["products"][0]["members"][0]["expname"])[1]
                    )[0].rsplit("_", 1)[0]
                    + "_drzl"
                )
                asn.new_product(product_name)
                new_members = asn.current_product["members"]
                if "_wfi_" in member["expname"]:
                    # Make and add a member for each detector
                    for det in WFI_DETECTORS:
                        new_member = copy.deepcopy(member)
                        new_member["expname"] = member["expname"].replace("wfi", det)
                        new_members.append(new_member)
            if asn.is_valid:
                results.append(asn)
                return results
            else:
                return None

    def _init_hook(self, item):
        """Post-check and pre-add initialization"""
        super()._init_hook(item)

        # Set which sequence counter should be used.
        self._sequence = self._sequences[self.data["asn_type"]]

        # Create the product.
        self.new_product()

        # Update meta data
        self.update_asn(item=item)

    def _add(self, item):
        """Add item to this association."""
        member = self.make_member(item)
        if self.is_member(member):
            # logger.debug(
            #     'Member is already part of the association:'
            #     '\n\tassociation: {}'
            #     '\n]tmember: {}'.format(self, member)
            # )
            return

        self.update_validity(member)
        members = self.current_product["members"]
        members.append(member)
        if member["exposerr"] not in _EMPTY:
            logger.warning(
                f"Member {item['filename']} has exposure error \"{member['exposerr']}\""
            )

        # Update meta info
        self.update_asn(item=item, member=member)

    def _add_items(self, items, product_name=None, with_exptype=False, **kwargs):
        """Force adding items to the association

        Parameters
        ----------
        items : [object[, ...]]
            A list of items to make members of the association.

        product_name : str or None
            The name of the product to add the items to.
            If the product does not already exist, it will be created.
            If None, the default DMS ELPP naming
            conventions will be attempted.

        with_exptype : bool
            If True, each item is expected to be a 2-tuple with
            the first element being the item to add as `expname`
            and the second items is the `exptype`

        kwargs : dict
            Allows other keyword arguments used by other subclasses.

        Notes
        -----
        This is a low-level shortcut into adding members, such as file names,
        to an association. All defined shortcuts and other initializations are
        by-passed, resulting in a potentially unusable association.
        """
        if product_name is None:
            raise AssociationNotValidError("Product name needs to be specified")

        self.new_product(product_name)
        members = self.current_product["members"]
        for item in items:
            exptype = "science"
            if with_exptype:
                item, exptype = item
            member = Member({"expname": item, "exptype": exptype}, item=item)
            self.update_validity(member)
            members.append(member)
        self.sequence = next(self._sequence)

    def __repr__(self):
        # flake8:  noqa: F821
        try:
            file_name, json_repr = self.ioregistry["json"].dump(self)
        except KeyValueRegistryError:
            return str(self.__class__)
        return json_repr

    def __str__(self):
        result_list = []
        result_list.append(
            f"{self.asn_name} with {len(self.data['products'])} products"
        )
        result_list.append(f"Rule={self.data['asn_rule']}")
        result_list.append(self.data["constraints"])
        result_list.append("Products:")
        for product in self.data["products"]:
            result_list.append(
                f"\t{product['name']} with {len(product['members'])} members"
            )
        result = "\n".join(result_list)
        return result


@RegistryMarker.utility
class Utility:
    """Utility functions that understand DMS Level 3 associations"""

    @staticmethod
    def resequence(associations):
        """Resequence the numbering for the ELPP association types"""
        counters = defaultdict(lambda: defaultdict(Counter))
        for asn in associations:
            asn.sequence = next(counters[asn.data["asn_id"]][asn.data["asn_type"]])

    @staticmethod
    def rename_to_level2(level1b_name, exp_type=None, member_exptype="science"):
        """Rename a Level 1b Exposure to a Level2 name.

        The basic transform is changing the suffix `uncal` to
        `cal`, `calints`, or `rate`.

        Parameters
        ----------
        level1b_name : str
            The Level 1b exposure name.

        exp_type:
            JWST exposure type. If not specified,
            it will be presumed that the name
            should get a Level2b name

        is_tso : boolean
            Use 'calints' instead of 'cal' as
            the suffix.

        member_exptype: str
            The association member exposure type, such as "science".

        Returns
        -------
        str
            The Level 2b name
        """
        match = re.match(_LEVEL1B_REGEX, level1b_name)
        if match is None or match.group("type") != "_uncal":
            logger.warning(
                (
                    'Item FILENAME="{}" is not a Level 1b name. Cannot transform to'
                    " Level 2b."
                ).format(level1b_name)
            )
            return level1b_name

        if member_exptype == "background":
            suffix = "x1d"
        else:
            if exp_type in LEVEL2B_EXPTYPES:
                suffix = "cal"
            else:
                suffix = "rate"

        # if is_tso:
        #    suffix += 'ints'

        level2_name = "".join(
            [match.group("path"), "_", suffix, match.group("extension")]
        )
        return level2_name

    @staticmethod
    def get_candidate_list(value):
        """Parse the candidate list from a item value

        Parameters
        ----------
        value : str
            The value from the item to parse. Usually
            item['ASN_CANDIDATE']

        Returns
        -------
        [ACID, ...]
            The list of parsed candidates.
        """

        result = []
        evaled = evaluate(value)
        if is_iterable(evaled):
            result = [ACID(v) for v in evaled]
        return result

    @staticmethod
    @RegistryMarker.callback("finalize")
    def finalize(associations):
        """Check validity and duplications in an association list

        Parameters
        ----------
        associations:[association[, ...]]
            List of associations

        Returns
        -------
        finalized_associations : [association[, ...]]
            The validated list of associations
        """
        finalized_asns = []
        lv3_asns = []
        for asn in associations:
            if isinstance(asn, DMS_ELPP_Base):
                finalized = asn.finalize()
                if finalized is not None:
                    lv3_asns.extend(finalized)
            else:
                finalized_asns.append(asn)

        lv3_asns = prune_duplicate_associations(lv3_asns)
        lv3_asns = prune_duplicate_products(lv3_asns)

        # Ensure sequencing is correct.
        Utility.resequence(lv3_asns)

        # Merge lists and return
        return finalized_asns + lv3_asns


# ---------
# Utilities
# ---------
# Define default product name filling
format_product = FormatTemplate(
    key_formats={"source_id": ["s{:05d}", "s{:s}"], "expspcin": ["{:0>2s}"]}
)


def dms_product_name_noopt(asn):
    """Define product name without any optical elements.

    Parameters
    ---------
    asn : Association
        The association for which the product
        name is to be created.

    Returns
    -------
    product_name : str
        The product name
    """
    target = asn._get_target()

    instrument = asn._get_instrument()

    product_name = "r{}-{}_{}_{}".format(
        #        asn.data['program'],
        asn.data["visi_id"],
        asn.acid.id,
        target,
        instrument,
    )
    return product_name.lower()


def dms_product_name_sources(asn):
    """Produce source-based product names

    Parameters
    ---------
    asn : Association
        The association for which the product
        name is to be created.

    Returns
    -------
    product_name : str
        The product name
    """
    instrument = asn._get_instrument()

    opt_elem = asn._get_opt_element()

    product_name_format = "r{program}-{acid}_{source_id}_{instrument}_{opt_elem}"
    product_name = format_product(
        product_name_format,
        #        program=asn.data['program'],
        program=asn.data["visit_id"],
        acid=asn.acid.id,
        instrument=instrument,
        opt_elem=opt_elem,
    )

    return product_name.lower()


# -----------------
# Basic constraints
# -----------------
class Constraint_Base(Constraint):
    """Select on program and instrument"""

    def __init__(self):
        super().__init__(
            [
                DMSAttrConstraint(
                    name="program",
                    sources=["program"],
                ),
                DMSAttrConstraint(
                    name="instrument",
                    sources=["instrume"],
                ),
            ],
            name="base",
        )


class Constraint_Instrument(Constraint):
    """Select on instrument"""

    def __init__(self):
        super().__init__(
            [
                DMSAttrConstraint(
                    name="instrument",
                    sources=["instrume"],
                ),
            ],
        )


class Constraint_Filename(DMSAttrConstraint):
    """Select on visit number"""

    def __init__(self):
        super().__init__(
            name="Filename",
            sources=["filename"],
        )


class Constraint_Expos(DMSAttrConstraint):
    """Select on exposure number"""

    def __init__(self):
        super().__init__(
            name="exposure_number",
            sources=["nexpsur"],
            force_unique=True,
            required=True,
        )


class Constraint_Tile(DMSAttrConstraint):
    """Select on exposure number"""

    def __init__(self):
        super().__init__(
            name="tile",
            sources=["tile"],
            force_unique=True,
            required=True,
        )


class Constraint_Image(DMSAttrConstraint):
    """Select on exposure type"""

    def __init__(self):
        super().__init__(
            name="exp_type",
            sources=["exp_type"],
            value="wfi_image|wfi_wfsc",
        )


class Constraint_Obsnum(DMSAttrConstraint):
    """Select on OBSNUM"""

    def __init__(self):
        super().__init__(
            name="obs_num",
            sources=["obs_num"],
            force_unique=False,
            required=False,
        )


class Constraint_Optical_Path(Constraint):
    """Select on optical path"""

    def __init__(self):
        super().__init__(
            [
                DMSAttrConstraint(
                    name="opt_elem",
                    sources=["opt_elem"],
                    required=False,
                )
            ]
        )


class Constraint_SubCategory(DMSAttrConstraint):
    """Select on SUBCAT (proposal subcategory) in the mock pool file"""

    def __init__(self):
        super().__init__(
            name="sub_cat",
            sources=["subcat"],
            force_unique=True,
            required=True,
        )


class Constraint_Category(DMSAttrConstraint):
    """Select on SUBCAT (proposal subcategory) in the mock pool file"""

    def __init__(self):
        super().__init__(
            name="category",
            sources=["cat"],
            force_unique=True,
            required=True,
        )


class Constraint_Pass(Constraint):
    """Select on pass number"""

    def __init__(self):
        super().__init__(
            [
                DMSAttrConstraint(
                    name="pass",
                    sources=["pass"],
                    required=True,
                    force_unique=True,
                )
            ],
            name="pass",
        )


class Constraint_Sequence(Constraint):
    """Select on pass number"""

    def __init__(self):
        super().__init__(
            [
                DMSAttrConstraint(
                    name="sequence",
                    sources=["sequence"],
                    required=True,
                    force_unique=True,
                )
            ],
            name="sequence",
        )


class Constraint_Image_Science(DMSAttrConstraint):
    """Select on science images"""

    def __init__(self):
        super().__init__(
            name="exp_type",
            sources=["exp_type"],
            value="|".join(IMAGE2_SCIENCE_EXP_TYPES),
        )


class Constraint_Single_Science(SimpleConstraint):
    """Allow only single science exposure

    Parameters
    ----------
    has_science_fn : func
        Function to determine whether the association
        has a science member already. No arguments are provided.

    sc_kwargs : dict
        Keyword arguments to pass to the parent class `SimpleConstraint`

    Notes
    -----
    The `has_science_fn` is further wrapped in a lambda function
    to provide a closure. Otherwise if the function is a bound method,
    that method may end up pointing to an instance that is not calling
    this constraint.
    """

    def __init__(self, has_science_fn, **sc_kwargs):
        super().__init__(
            name="single_science",
            value=False,
            sources=lambda item: has_science_fn(),
            **sc_kwargs,
        )


class Constraint_Spectral(DMSAttrConstraint):
    """Constrain on spectral exposure types"""

    def __init__(self):
        super().__init__(
            name="exp_type",
            sources=["exp_type"],
            value="wfi_grism|wfi_prism",
            force_unique=False,
        )


class Constraint_Spectral_Science(Constraint):
    """Select on spectral science

    Parameters
    exclude_exp_types : [exp_type[, ...]]
        List of exposure types to not consider from
        from the general list.
    """

    def __init__(self, exclude_exp_types=None):
        if exclude_exp_types is None:
            general_science = SPEC2_SCIENCE_EXP_TYPES
        else:
            general_science = set(SPEC2_SCIENCE_EXP_TYPES).symmetric_difference(
                exclude_exp_types
            )

        super().__init__(
            [
                DMSAttrConstraint(
                    name="exp_type",
                    sources=["exp_type"],
                    value="|".join(general_science),
                )
            ],
            reduce=Constraint.any,
        )


class Constraint_Target(DMSAttrConstraint):
    """Select on target

    Parameters
    ----------
    association: Association
        If specified, use the `get_exposure_type` method
        to as part of the target selection.
    """

    def __init__(self, association=None):
        if association is None:
            super().__init__(
                name="target",
                sources=["targname"],
            )
        else:
            super().__init__(
                name="target",
                sources=["targname"],
                onlyif=lambda item: association.get_exposure_type(item) != "background",
                force_reprocess=ProcessList.EXISTING,
                only_on_match=True,
            )


# -----------
# Base Mixins
# -----------
class AsnMixin_AuxData:
    """Process special and non-science exposures as science."""

    def get_exposure_type(self, item, default="science"):
        """Override to force exposure type to always be science
        Parameters
        ----------
        item : dict
            The pool entry for which the exposure type is determined
        default : str or None
            The default exposure type.
            If None, routine will raise LookupError
        Returns
        -------
        exposure_type : 'science'
            Returns as science for most Exposures
        exposure_type : 'target_acquisition'
            Returns target_acquisition for mir_tacq
        """
        NEVER_CHANGE = ["target_acquisition"]
        exp_type = super().get_exposure_type(item, default=default)
        if exp_type in NEVER_CHANGE:
            return exp_type
        return "science"


class AsnMixin_Science(DMS_ELPP_Base):
    """Basic science constraints"""

    def __init__(self, *args, **kwargs):
        # Setup target acquisition inclusion
        constraint_acqs = Constraint(
            [
                DMSAttrConstraint(
                    name="acq_obsnum",
                    sources=["obs_num"],
                    value=lambda: "("
                    + "|".join(self.constraints["obs_num"].found_values)
                    + ")",
                    force_unique=False,
                )
            ],
            name="acq_constraint",
            work_over=ProcessList.EXISTING,
        )

        # Put all constraints together.
        self.constraints = Constraint(
            [
                Constraint_Base(),
                DMSAttrConstraint(sources=["is_imprt"], force_undefined=True),
                Constraint(
                    [
                        Constraint(
                            [self.constraints, Constraint_Obsnum()], name="rule"
                        ),
                        constraint_acqs,
                    ],
                    name="acq_check",
                    reduce=Constraint.any,
                ),
            ],
            name="dmsbase_top",
        )

        super().__init__(*args, **kwargs)


class AsnMixin_Spectrum(AsnMixin_Science):
    """All things that are spectrum"""

    def _init_hook(self, item):
        """Post-check and pre-add initialization"""

        self.data["asn_type"] = "spec3"
        super()._init_hook(item)


# ---------------------------------------------
# Mixins to define the broad category of rules.
# ---------------------------------------------
class AsnMixin_Lv2FOV:
    """Level 2 Image association base"""

    def _init_hook(self, item):
        """Post-check and pre-add initialization"""

        super()._init_hook(item)
        self.data["asn_type"] = "FOV"

    def finalize(self):
        """Finalize association


        Returns
        -------
        associations: [association[, ...]] or None
            List of fully-qualified associations that this association
            represents.
            `None` if a complete association cannot be produced.

        """
        if self.is_valid:
            return self.make_fov_asn()
        else:
            return None


class AsnMixin_Lv2Image:
    """Level 2 Image association base"""

    def _init_hook(self, item):
        """Post-check and pre-add initialization"""

        super()._init_hook(item)
        self.data["asn_type"] = "image"


class AsnMixin_Lv2GBTDSpass:
    """Level 2 GBTDS association base"""

    def _init_hook(self, item):
        """Post-check and pre-add initialization"""

        super()._init_hook(item)
        self.data["asn_type"] = "pass"


class AsnMixin_Lv2GBTDSfull:
    """Level 2 GBTDS association base"""

    def _init_hook(self, item):
        """Post-check and pre-add initialization"""

        super()._init_hook(item)
        self.data["asn_type"] = "full"



"""Suffix manipulation

API
---
`remove_suffix`
    Main function used to remove suffixes from
    file base names.

Notes
-----

`KNOW_SUFFIXES` is the list used by `remove_suffix`. This
list is generated by the function `combine_suffixes`. The function uses
`SUFFIXES_TO_ADD` to add other suffixes that it would otherwise not
discover or should absolutely be in the list. The function uses
'SUFFIXES_TO_DISCARD` for strings found that are not to be considered
suffixes.

Hence, to update `KNOW_SUFFIXES`, update both `SUFFIXES_TO_ADD` and
`SUFFIXES_TO_DISCARD` as necessary, then use the output of
`find_suffixes`.
"""

import itertools
import logging
import re

__all__ = ["remove_suffix"]

# Configure logging
logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())

# Suffixes that are hard-coded or otherwise
# have to exist. Used by `find_suffixes` to
# add to the result it produces.
SUFFIXES_TO_ADD = [
    "ca",
    "crf",
    "cal",
    "dark",
    "flat",
    "median",
    "phot",
    "photom",
    "ramp",
    "rate",
    "uncal",
    "assignwcs",
    "dq_init",
    "dqinit",
    "assign_wcs",
    "linearity",
    "jump",
    "rampfit",
    "saturation",
    "sourcedetection",
    "dark_current",
    "darkcurrent",
    "outlier_detection",
    "skymatch",
    "refpix",
    "cat",
    "segm",
    "i2d",
]

# Suffixes that are discovered but should not be considered.
# Used by `find_suffixes` to remove undesired values it has found.
SUFFIXES_TO_DISCARD = ["highlevelpipeline", "pipeline", "step"]


# Calculated suffixes.
_calculated_suffixes = {
    "saturationstep",
    "darkcurrentstep",
    "jumpstep",
    "rampfit",
    "dark_current",
    "assignwcsstep",
    "flatfieldstep",
    "fluxstep",
    "step",
    "dqinitstep",
    "assign_wcs",
    "linearity",
    "rampfitstep",
    "photomstep",
    "pipeline",
    "dq_init",
    "linearitystep",
    "dark_current",
    "jump",
    "sourcedetection",
    "sourcedetectionstep",
    "tweakregstep",
    "outlierdetectionstep",
    "skymatchstep",
    "refpixstep",
    "resamplestep",
    "sourcecatalogstep",
}


# ##########
# Suffix API
# ##########
def remove_suffix(name):
    """Remove the suffix if a known suffix is already in name"""
    separator = None
    match = REMOVE_SUFFIX_REGEX.match(name)
    try:
        name = match.group("root")
        separator = match.group("separator")
    except AttributeError:
        pass
    if separator is None:
        separator = "_"
    return name, separator


def replace_suffix(name, new_suffix):
    """Replace suffix on name

    Parameters
    ----------
    name: str
        The name to replace the suffix of.
        Expected to be only the basename; no extensions.

    new_suffix:
        The new suffix to use.
    """
    no_suffix, separator = remove_suffix(name)
    return no_suffix + separator + new_suffix


# #####################################
# Functions to generate `KNOW_SUFFIXES`
# #####################################
def combine_suffixes(
    to_add=(_calculated_suffixes, SUFFIXES_TO_ADD), to_remove=(SUFFIXES_TO_DISCARD,)
):
    """Combine the suffix lists into a single list

    Parameters
    ----------
    to_add: [iterable[, ...]]
        List of iterables to add to the combined list.

    to_remove: [iterable[, ...]]
        List of iterables to remove from the combined list.

    Returns
    -------
    suffixes: list
        The list of suffixes.
    """
    combined = set(itertools.chain.from_iterable(to_add))
    combined.difference_update(itertools.chain.from_iterable(to_remove))
    combined = list(combined)
    combined.sort()

    return combined


def find_suffixes():
    """Find all possible suffixes from the romancal package

    Returns
    -------
    suffixes: set
        The set of all programmatically findable suffixes.

    Notes
    -----
    This will load all of the `romancal` package. Consider if this
    is worth doing dynamically or only as a utility to update
    a static list.
    """
    from romancal.stpipe.utilities import all_steps

    # First traverse the code base and find all
    # `Step` classes. The default suffix is the
    # class name.
    suffixes = {klass_name.lower() for klass_name, klass in all_steps().items()}

    # That's all folks
    return list(suffixes)


# --------------------------------------------------
# The set of suffixes used by the pipeline.
# This set is generated by `combine_suffixes`.
# Only update this list by `combine_suffixes`.
# Modify `SUFFIXES_TO_ADD` and `SUFFIXES_TO_DISCARD`
# to change the results.
# --------------------------------------------------
KNOW_SUFFIXES = combine_suffixes()

# Regex for removal
REMOVE_SUFFIX_REGEX = re.compile(
    "^(?P<root>.+?)((?P<separator>_|-)(" + "|".join(KNOW_SUFFIXES) + "))?$"
)


# ############################################
# Main
# Find and report differences from known list.
# ############################################
if __name__ == "__main__":
    print("Searching code base for calibration suffixes...")
    calculated_suffixes = find_suffixes()
    found_suffixes = combine_suffixes(
        to_add=(calculated_suffixes, SUFFIXES_TO_ADD), to_remove=(SUFFIXES_TO_DISCARD,)
    )
    print(
        "Known list has {known_len} suffixes. Found {new_len} suffixes.".format(
            known_len=len(KNOW_SUFFIXES), new_len=len(found_suffixes)
        )
    )
    print(
        "Suffixes that have changed are"
        f" {set(found_suffixes).symmetric_difference(KNOW_SUFFIXES)}"
    )



"""Primary code for performing outlier detection on Roman observations."""

import logging
import warnings
from functools import partial

import numpy as np
from astropy.stats import sigma_clip
from astropy.units import Quantity
from drizzle.cdrizzle import tblot
from roman_datamodels import datamodels as rdm
from roman_datamodels.dqflags import pixel
from scipy import ndimage

from romancal.datamodels import ModelContainer
from romancal.resample import resample
from romancal.resample.resample_utils import build_driz_weight, calc_gwcs_pixmap

from ..stpipe import RomanStep

log = logging.getLogger(__name__)
log.setLevel(logging.DEBUG)

__all__ = ["OutlierDetection", "flag_cr", "abs_deriv"]


class OutlierDetection:
    """Main class for performing outlier detection.

    This is the controlling routine for the outlier detection process.
    It loads and sets the various input data and parameters needed by
    the various functions and then controls the operation of this process
    through all the steps used for the detection.

    Notes
    -----
    This routine performs the following operations::

      1. Extracts parameter settings from input model and merges
         them with any user-provided values
      2. Resamples all input images into grouped observation mosaics.
      3. Creates a median image from all grouped observation mosaics.
      4. Blot median image to match each original input image.
      5. Perform statistical comparison between blotted image and original
         image to identify outliers.
      6. Updates input data model DQ arrays with mask of detected outliers.

    """

    default_suffix = "i2d"

    def __init__(self, input_models, **pars):
        """
        Initialize the class with input ModelContainers.

        Parameters
        ----------
        input_models : ~romancal.datamodels.container.ModelContainer
            A `~romancal.datamodels.container.ModelContainer` object containing the data
            to be processed.

        pars : dict, optional
            Optional user-specified parameters to modify how outlier_detection
            will operate.  Valid parameters include:
            - resample_suffix

        """
        self.input_models = input_models

        self.outlierpars = dict(pars)
        self.resample_suffix = f"_outlier_{self.default_suffix if pars.get('resample_suffix') is None else pars.get('resample_suffix')}.asdf"
        log.debug(f"Defined output product suffix as: {self.resample_suffix}")

        # Define how file names are created
        self.make_output_path = pars.get(
            "make_output_path", partial(RomanStep._make_output_path, None)
        )

    def do_detection(self):
        """Flag outlier pixels in DQ of input images."""  # self._convert_inputs()
        pars = self.outlierpars

        if pars["resample_data"]:
            # Start by creating resampled/mosaic images for
            # each group of exposures
            resamp = resample.ResampleData(
                self.input_models, single=False, blendheaders=False, **pars
            )
            drizzled_models = resamp.do_drizzle()

        else:
            # for non-dithered data, the resampled image is just the original image
            drizzled_models = self.input_models
            for model in drizzled_models:
                model["weight"] = build_driz_weight(
                    model,
                    weight_type="ivm",
                    good_bits=pars["good_bits"],
                )

        # Initialize intermediate products used in the outlier detection
        median_model = (
            rdm.open(drizzled_models[0]).copy()
            if isinstance(drizzled_models[0], str)
            else drizzled_models[0].copy()
        )

        # Perform median combination on set of drizzled mosaics
        median_model.data = Quantity(
            self.create_median(drizzled_models), unit=median_model.data.unit
        )

        if not pars.get("in_memory", True):
            median_model.meta.filename = "drizzled_median.asdf"
            median_model_output_path = self.make_output_path(
                basepath=median_model.meta.filename,
                suffix="median",
            )
            median_model.save(median_model_output_path)
            log.info(f"Saved model in {median_model_output_path}")

        if pars["resample_data"]:
            # Blot the median image back to recreate each input image specified
            # in the original input list/ASN/ModelContainer
            blot_models = self.blot_median(median_model)

        else:
            # Median image will serve as blot image
            blot_models = ModelContainer(return_open=False)
            for _ in range(len(self.input_models)):
                blot_models.append(median_model)

        # Perform outlier detection using statistical comparisons between
        # each original input image and its blotted version of the median image
        self.detect_outliers(blot_models)

        # clean-up (just to be explicit about being finished with
        # these results)
        del median_model, blot_models

    def create_median(self, resampled_models):
        """Create a median image from the singly resampled images.

        NOTES
        -----
        This version is simplified from astrodrizzle's version in the
        following ways:
        - type of combination: fixed to 'median'
        - 'minmed' not implemented as an option
        """
        maskpt = self.outlierpars.get("maskpt", 0.7)

        log.info("Computing median")

        # Compute weight means without keeping DataModel for eacn input open
        # Start by insuring that the ModelContainer does NOT open and keep each datamodel
        ropen_orig = resampled_models._return_open
        resampled_models._return_open = False  # turn off auto-opening of models
        # keep track of resulting computation for each input resampled datamodel
        weight_thresholds = []
        # For each model, compute the bad-pixel threshold from the weight arrays
        for resampled in resampled_models:
            m = rdm.open(resampled)
            weight = m.weight
            # necessary in order to assure that mask gets applied correctly
            if hasattr(weight, "_mask"):
                del weight._mask
            mask_zero_weight = np.equal(weight, 0.0)
            mask_nans = np.isnan(weight)
            # Combine the masks
            weight_masked = np.ma.array(
                weight, mask=np.logical_or(mask_zero_weight, mask_nans)
            )
            # Sigma-clip the unmasked data
            weight_masked = sigma_clip(weight_masked, sigma=3, maxiters=5)
            mean_weight = np.mean(weight_masked)
            # Mask pixels where weight falls below maskpt percent
            weight_threshold = mean_weight * maskpt
            weight_thresholds.append(weight_threshold)
            # close and delete the model, just to explicitly try to keep the memory as clean as possible
            m.close()
            del m
        # Reset ModelContainer attribute to original value
        resampled_models._return_open = ropen_orig

        # Now, set up buffered access to all input models
        resampled_models.set_buffer(1.0)  # Set buffer at 1Mb
        resampled_sections = resampled_models.get_sections()
        median_image = np.empty(
            (resampled_models.imrows, resampled_models.imcols),
            resampled_models.imtype,
        )
        median_image[:] = np.nan  # initialize with NaNs

        for resampled_sci, resampled_weight, (row1, row2) in resampled_sections:
            # Create a mask for each input image, masking out areas where there is
            # no data or the data has very low weight
            badmasks = []
            for weight, weight_threshold in zip(resampled_weight, weight_thresholds):
                badmask = np.less(weight, weight_threshold)
                log.debug(
                    f"Percentage of pixels with low weight: {np.sum(badmask) / len(weight.flat) * 100}"
                )
                badmasks.append(badmask)

            # Fill resampled_sci array with nan's where mask values are True
            for f1, f2 in zip(resampled_sci, badmasks):
                for elem1, elem2 in zip(f1, f2):
                    elem1[elem2] = np.nan

            del badmasks

            # For a stack of images with "bad" data replaced with Nan
            # use np.nanmedian to compute the median.
            with warnings.catch_warnings():
                warnings.filterwarnings(
                    action="ignore",
                    message="All-NaN slice encountered",
                    category=RuntimeWarning,
                )
                median_image[row1:row2] = np.nanmedian(resampled_sci, axis=0)
            del resampled_sci, resampled_weight

        return median_image

    def blot_median(self, median_model):
        """Blot resampled median image back to the detector images."""
        interp = self.outlierpars.get("interp", "linear")
        sinscl = self.outlierpars.get("sinscl", 1.0)
        in_memory = self.outlierpars.get("in_memory", True)

        # Initialize container for output blot images
        blot_models = []

        log.info("Blotting median")
        for model in self.input_models:
            blotted_median = model.copy()

            # clean out extra data not related to blot result
            blotted_median.err *= 0.0  # None
            blotted_median.dq *= 0  # None

            # apply blot to re-create model.data from median image
            blotted_median.data = Quantity(
                gwcs_blot(median_model, model, interp=interp, sinscl=sinscl),
                unit=blotted_median.data.unit,
            )
            if not in_memory:
                model_path = self.make_output_path(
                    basepath=model.meta.filename, suffix="blot"
                )
                blotted_median.save(model_path)
                log.info(f"Saved model in {model_path}")

                # Append model name to the ModelContainer so it is not passed in memory
                blot_models.append(model_path)
            else:
                blot_models.append(blotted_median)

        return ModelContainer(blot_models, return_open=in_memory)

    def detect_outliers(self, blot_models):
        """Flag DQ array for cosmic rays in input images.

        The science frame in each ImageModel in self.input_models is compared to
        the corresponding blotted median image in blot_models. The result is
        an updated DQ array in each ImageModel in input_models.

        Parameters
        ----------
        blot_models : JWST ModelContainer object
            data model container holding ImageModels of the median output frame
            blotted back to the wcs and frame of the ImageModels in
            input_models

        Returns
        -------
        None
            The dq array in each input model is modified in place

        """
        log.info("Flagging outliers")
        for i, (image, blot) in enumerate(zip(self.input_models, blot_models)):
            blot = rdm.open(blot)
            flag_cr(image, blot, **self.outlierpars)
            self.input_models[i] = image


def flag_cr(
    sci_image,
    blot_image,
    snr="5.0 4.0",
    scale="1.2 0.7",
    backg=0,
    resample_data=True,
    **kwargs,
):
    """Masks outliers in science image by updating DQ in-place

    Mask blemishes in dithered data by comparing a science image
    with a model image and the derivative of the model image.

    Parameters
    ----------
    sci_image : ~romancal.DataModel.ImageModel
        the science data

    blot_image : ~romancal.DataModel.ImageModel
        the blotted median image of the dithered science frames

    snr : str
        Signal-to-noise ratio

    scale : str
        scaling factor applied to the derivative

    backg : float
        Background value (scalar) to subtract

    resample_data : bool
        Boolean to indicate whether blot_image is created from resampled,
        dithered data or not
    """
    snr1, snr2 = (float(val) for val in snr.split())
    scale1, scale2 = (float(val) for val in scale.split())

    # Get background level of science data if it has not been subtracted, so it
    # can be added into the level of the blotted data, which has been
    # background-subtracted
    if (
        hasattr(sci_image.meta, "background")
        and sci_image.meta.background.subtracted is False
        and sci_image.meta.background.level is not None
    ):
        subtracted_background = sci_image.meta.background.level
        log.debug(f"Adding background level {subtracted_background} to blotted image")
    else:
        # No subtracted background.  Allow user-set value, which defaults to 0
        subtracted_background = backg

    sci_data = sci_image.data
    blot_data = blot_image.data
    blot_deriv = abs_deriv(blot_data.value)
    err_data = np.nan_to_num(sci_image.err)

    # create the outlier mask
    if resample_data:  # dithered outlier detection
        blot_data += subtracted_background
        diff_noise = np.abs(sci_data - blot_data)

        # Create a boolean mask based on a scaled version of
        # the derivative image (dealing with interpolating issues?)
        # and the standard n*sigma above the noise
        threshold1 = scale1 * blot_deriv + snr1 * err_data.value
        mask1 = np.greater(diff_noise.value, threshold1)

        # Smooth the boolean mask with a 3x3 boxcar kernel
        kernel = np.ones((3, 3), dtype=int)
        mask1_smoothed = ndimage.convolve(mask1, kernel, mode="nearest")

        # Create a 2nd boolean mask based on the 2nd set of
        # scale and threshold values
        threshold2 = scale2 * blot_deriv + snr2 * err_data.value
        mask2 = np.greater(diff_noise.value, threshold2)

        # Final boolean mask
        cr_mask = mask1_smoothed & mask2

    else:  # stack outlier detection
        diff_noise = np.abs(sci_data - blot_data)

        # straightforward detection of outliers for non-dithered data since
        # err_data includes all noise sources (photon, read, and flat for baseline)
        cr_mask = np.greater(diff_noise.value, snr1 * err_data.value)

    # Count existing DO_NOT_USE pixels
    count_existing = np.count_nonzero(sci_image.dq & pixel.DO_NOT_USE)

    # Update the DQ array values in the input image but preserve datatype.
    sci_image.dq = np.bitwise_or(
        sci_image.dq, cr_mask * (pixel.DO_NOT_USE | pixel.OUTLIER)
    ).astype(np.uint32)

    # Report number (and percent) of new DO_NOT_USE pixels found
    count_outlier = np.count_nonzero(sci_image.dq & pixel.DO_NOT_USE)
    count_added = count_outlier - count_existing
    percent_cr = count_added / (sci_image.shape[0] * sci_image.shape[1]) * 100
    log.info(f"New pixels flagged as outliers: {count_added} ({percent_cr:.2f}%)")


def abs_deriv(array):
    """Take the absolute derivative of a numpy array."""
    tmp = np.zeros(array.shape, dtype=np.float64)
    out = np.zeros(array.shape, dtype=np.float64)

    tmp[1:, :] = array[:-1, :]
    tmp, out = _absolute_subtract(array, tmp, out)
    tmp[:-1, :] = array[1:, :]
    tmp, out = _absolute_subtract(array, tmp, out)

    tmp[:, 1:] = array[:, :-1]
    tmp, out = _absolute_subtract(array, tmp, out)
    tmp[:, :-1] = array[:, 1:]
    tmp, out = _absolute_subtract(array, tmp, out)

    return out


def _absolute_subtract(array, tmp, out):
    tmp = np.abs(array - tmp)
    out = np.maximum(tmp, out)
    tmp = tmp * 0.0
    return tmp, out


def gwcs_blot(median_model, blot_img, interp="poly5", sinscl=1.0):
    """
    Resample the output/resampled image to recreate an input image based on
    the input image's world coordinate system

    Parameters
    ----------
    median_model : `~roman_datamodels.datamodels.MosaicModel`

    blot_img : datamodel
        Datamodel containing header and WCS to define the 'blotted' image

    interp : str, optional
        The type of interpolation used in the resampling. The
        possible values are "nearest" (nearest neighbor interpolation),
        "linear" (bilinear interpolation), "poly3" (cubic polynomial
        interpolation), "poly5" (quintic polynomial interpolation),
        "sinc" (sinc interpolation), "lan3" (3rd order Lanczos
        interpolation), and "lan5" (5th order Lanczos interpolation).

    sinscl : float, optional
        The scaling factor for sinc interpolation.
    """
    blot_wcs = blot_img.meta.wcs

    # Compute the mapping between the input and output pixel coordinates
    pixmap = calc_gwcs_pixmap(blot_wcs, median_model.meta.wcs, blot_img.data.shape)
    log.debug(f"Pixmap shape: {pixmap[:, :, 0].shape}")
    log.debug(f"Sci shape: {blot_img.data.shape}")

    pix_ratio = 1
    log.info(f"Blotting {blot_img.data.shape} <-- {median_model.data.shape}")

    outsci = np.zeros(blot_img.shape, dtype=np.float32)

    # Currently tblot cannot handle nans in the pixmap, so we need to give some
    # other value.  -1 is not optimal and may have side effects.  But this is
    # what we've been doing up until now, so more investigation is needed
    # before a change is made.  Preferably, fix tblot in drizzle.
    pixmap[np.isnan(pixmap)] = -1
    tblot(
        median_model.data,
        pixmap,
        outsci,
        scale=pix_ratio,
        kscale=1.0,
        interp=interp,
        exptime=1.0,
        misval=0.0,
        sinscl=sinscl,
    )

    return outsci



import copy
import json
import os
from io import StringIO
from pathlib import Path
from typing import Tuple

import numpy as np
import pytest
import requests
from astropy import coordinates as coord
from astropy import table
from astropy import units as u
from astropy.modeling import models
from astropy.modeling.models import RotationSequence3D, Scale, Shift
from astropy.table import Table
from astropy.time import Time
from gwcs import coordinate_frames as cf
from gwcs import wcs
from gwcs.geometry import CartesianToSpherical, SphericalToCartesian
from roman_datamodels import datamodels as rdm
from roman_datamodels import maker_utils

from romancal.datamodels import ModelContainer
from romancal.tweakreg import tweakreg_step as trs
from romancal.tweakreg.astrometric_utils import get_catalog


class MockConnectionError:
    def __init__(self, *args, **kwargs):
        raise requests.exceptions.ConnectionError


@pytest.fixture()
def test_data_dir():
    return Path.joinpath(Path(__file__).parent, "data")


def create_custom_catalogs(tmp_path, base_image, catalog_format="ascii.ecsv"):
    """
    Creates a custom catalog with three datamodels and its respective catalog data.
    """
    img1 = base_image(shift_1=1000, shift_2=1000)
    img2 = base_image(shift_1=1010, shift_2=1010)
    img3 = base_image(shift_1=1020, shift_2=1020)

    img1.meta.filename = "img1.asdf"
    img2.meta.filename = "img2.asdf"
    img3.meta.filename = "img3.asdf"

    # create valid custom catalog data to be used with each input datamodel
    catalog_data1 = get_catalog_data(img1)
    catalog_data2 = get_catalog_data(img2)
    catalog_data3 = get_catalog_data(img3)

    custom_catalog_map = [
        {
            "cat_filename": "ref_catalog_1",
            "cat_datamodel": img1.meta.filename,
            "cat_data": catalog_data1,
        },
        {
            "cat_filename": "ref_catalog_2",
            "cat_datamodel": img2.meta.filename,
            "cat_data": catalog_data2,
        },
        {
            "cat_filename": "ref_catalog_3",
            "cat_datamodel": img3.meta.filename,
            "cat_data": catalog_data3,
        },
    ]

    # create catfile
    catfile = str(tmp_path / "catfile.txt")
    catfile_content = StringIO()
    for x in custom_catalog_map:
        # write line to catfile
        catfile_content.write(f"{x.get('cat_datamodel')} {x.get('cat_filename')}\n")
        # write out the catalog data
        t = table.Table(x.get("cat_data"), names=("x", "y"))
        t.write(tmp_path / x.get("cat_filename"), format=catalog_format)
    with open(catfile, mode="w") as f:
        print(catfile_content.getvalue(), file=f)

    return {"catfile": catfile, "datamodels": [img1, img2, img3]}


def create_asn_file(tmp_path, members_mapping=None):
    asn_content = """
        {
            "asn_type": "None",
            "asn_rule": "DMS_ELPP_Base",
            "version_id": null,
            "code_version": "0.9.1.dev28+ge987cc9.d20230106",
            "degraded_status": "No known degraded exposures in association.",
            "program": "noprogram",
            "constraints": "No constraints",
            "asn_id": "a3001",
            "target": "none",
            "asn_pool": "test_pool_name",
            "products": [
                {
                    "name": "files.asdf",
                    "members": [
                        {
                            "expname": "img_1.asdf",
                            "exptype": "science"
                        },
                        {
                            "expname": "img_2.asdf",
                            "exptype": "science"
                        }
                    ]
                }
            ]
        }
    """
    if members_mapping is not None:
        asn_dict = json.loads(asn_content)
        asn_dict["products"][0]["members"] = []
        for x in members_mapping:
            asn_dict["products"][0]["members"].append(x)
        asn_content = json.dumps(asn_dict)

    asn_file_path = str(tmp_path / "sample_asn.json")
    asn_file = StringIO()
    asn_file.write(asn_content)
    with open(asn_file_path, mode="w") as f:
        print(asn_file.getvalue(), file=f)

    return asn_file_path


def update_wcsinfo(input_dm):
    """
    Update WCSInfo with realistic data (i.e. obtained from romanisim simulations).

    Parameters
    ----------
    input_dm : roman_datamodels.ImageModel
        A Roman image datamodel.
    """
    input_dm.meta.wcsinfo.v2_ref = 0.42955128282521254
    input_dm.meta.wcsinfo.v3_ref = -0.2479976768255853
    input_dm.meta.wcsinfo.vparity = -1
    input_dm.meta.wcsinfo.v3yangle = -999999
    input_dm.meta.wcsinfo.ra_ref = 270.0
    input_dm.meta.wcsinfo.dec_ref = 66.0
    input_dm.meta.wcsinfo.roll_ref = 60
    input_dm.meta.wcsinfo.s_region = (
        "POLYGON ICRS "
        "269.3318903230621 65.56866666048172 "
        "269.32578768154605 65.69246311613287 "
        "269.02457173246125 65.69201346248587 "
        "269.0333096074621 65.56870823657276 "
    )


def _create_tel2sky_model(input_dm):
    """
    Create the transform from telescope to sky.

    The transform is defined with a reference point in a Frame
    associated with the telescope (V2, V3) in arcsec, the corresponding
    reference point on sky (RA_REF, DEC_REF) in deg, and the position angle
    at the center of the aperture, ROLL_REF in deg.

    Parameters
    ----------
    input_dm : roman_datamodels.ImageModel
        A Roman image datamodel.

    Returns
    -------
    astropy.modeling.core.CompoundModel
        An astropy compound model with the transformations to map
        the telescope reference system onto the world reference systems.
    """
    v2_ref = input_dm.meta.wcsinfo.v2_ref / 3600
    v3_ref = input_dm.meta.wcsinfo.v3_ref / 3600
    roll_ref = input_dm.meta.wcsinfo.roll_ref
    ra_ref = input_dm.meta.wcsinfo.ra_ref
    dec_ref = input_dm.meta.wcsinfo.dec_ref

    angles = np.array([v2_ref, -v3_ref, roll_ref, dec_ref, -ra_ref])
    axes = "zyxyz"
    rot = RotationSequence3D(angles, axes_order=axes)

    # The sky rotation expects values in deg.
    # This should be removed when models work with quantities.
    model = (
        (Scale(0.1 / 3600) & Scale(0.1 / 3600))
        | SphericalToCartesian(wrap_lon_at=180)
        | rot
        | CartesianToSpherical(wrap_lon_at=360)
    )
    model.name = "v23tosky"
    return model


def create_basic_wcs(
    img_shape: tuple = (100, 100),
    ref_pix: tuple = (0, 0),
    ref_val: Tuple[u.Quantity, u.Quantity] = (
        u.Quantity("10 deg"),
        u.Quantity("0 deg"),
    ),
    pix_scale: Tuple[u.Quantity, u.Quantity] = (
        u.Quantity("0.1 arcsec"),
        u.Quantity("0.1 arcsec"),
    ),
    theta: u.Quantity = u.Quantity("0 deg"),
):
    """
    Creates a basic WCS (no distortion) to map pixel coordinates
    onto celestial spherical coordinates.

    Parameters
    ----------
    img_shape : tuple, optional
        Dimension of the pixel coordinates, by default (100, 100).
    ref_pix : tuple, optional
        Projection plane coordinates of the reference pixel, by default (0, 0)
    ref_val : tuple[u.Quantity, u.Quantity], optional
        Celestial coordinates at the reference pixel in degrees,
        by default (10, 0)*u.degree
    pix_scale : tuple[u.Quantity, u.Quantity], optional
        Size of a pixel side in arcsec, by default (0.1, 0.1)*u.arcsec
    theta : u.Quantity, optional
        CCW rotation angle of the projection plane, by default 0*u.deg

    Returns
    -------
    type: gwcs.wcs.WCS object
        A gwcs.wcs.WCS object with a model f : (x, y) |-> (ra, dec).

    Notes
    -----
    With the default parameters, the result will be a WCS for an image
    that's 100x100 pixels (@0.1"/pix in both axis) and reference
    coordinates (ra=10 deg, dec=0 deg, ICRS) sitting at the origin of the
    pixel coordinates frame (0,0).

    The WCS object created by this method generates a pipeline that
    does not contain the required steps to validate against the
    TweakReg pipeline.
    """

    # linear transformations
    shift_pixel_coords = models.Shift(-ref_pix[0]) & models.Shift(-ref_pix[1])
    scaling_matrix = np.array(
        [[pix_scale.to("deg")[0].value, 0], [0, pix_scale.to("deg")[1].value]]
    )
    rot_matrix = np.array(
        [
            [np.cos(theta), -np.sin(theta)],
            [np.sin(theta), np.cos(theta)],
        ]
    )
    matrix = np.dot(rot_matrix, scaling_matrix)
    affine_transf = models.AffineTransformation2D(matrix)
    affine_transf.inverse = models.AffineTransformation2D(np.linalg.inv(matrix))
    tan = models.Pix2Sky_TAN()
    celestial_rotation = models.RotateNative2Celestial(
        ref_val[0].value, ref_val[1].value, 180
    )
    det2sky = shift_pixel_coords | affine_transf | tan | celestial_rotation
    det2sky.name = "linear_transform"

    # frames
    detector_frame = cf.Frame2D(
        name="detector", unit=(u.pix, u.pix), axes_names=("x", "y")
    )
    cf.Frame2D(
        name="v2v3",
        axes_order=(0, 1),
        axes_names=("v2", "v3"),
        unit=(u.arcsec, u.arcsec),
    )
    sky_frame = cf.CelestialFrame(
        name="sky", unit=(u.deg, u.deg), reference_frame=coord.ICRS()
    )

    pipeline = [(detector_frame, det2sky), (sky_frame, None)]
    wcsobj = wcs.WCS(pipeline)
    wcsobj.bounding_box = (
        (-0.5, img_shape[0] + 0.5),
        (-0.5, img_shape[1] + 0.5),
    )

    return wcsobj


def create_wcs_for_tweakreg_pipeline(input_dm, shift_1=0, shift_2=0):
    """
    Create a basic WCS object (with optional shift) and
    append it to the input_dm.meta attribute.

    The WCS object will have a pipeline with the required
    steps to validate against the TweakReg pipeline.

    Parameters
    ----------
    input_dm : roman_datamodels.ImageModel
        A Roman image datamodel.
    shift_1 : int, optional
        Shift to be applied in the direction of the first axis, by default 0
    shift_2 : int, optional
        Shift to be applied in the direction of the second axis, by default 0
    """

    shape = input_dm.data.shape

    # create necessary transformations
    distortion = Shift(-shift_1) & Shift(-shift_2)
    distortion.bounding_box = ((-0.5, shape[-1] + 0.5), (-0.5, shape[-2] + 0.5))
    tel2sky = _create_tel2sky_model(input_dm)

    # create required frames
    detector = cf.Frame2D(name="detector", axes_order=(0, 1), unit=(u.pix, u.pix))
    v2v3 = cf.Frame2D(
        name="v2v3",
        axes_order=(0, 1),
        axes_names=("v2", "v3"),
        unit=(u.arcsec, u.arcsec),
    )
    world = cf.CelestialFrame(reference_frame=coord.ICRS(), name="world")

    # create pipeline
    pipeline = [
        wcs.Step(detector, distortion),
        wcs.Step(v2v3, tel2sky),
        wcs.Step(world, None),
    ]

    wcs_obj = wcs.WCS(pipeline)

    input_dm.meta["wcs"] = wcs_obj


def get_catalog_data(input_dm):
    gaia_cat = get_catalog(ra=270, dec=66, sr=100 / 3600)
    gaia_source_coords = [(ra, dec) for ra, dec in zip(gaia_cat["ra"], gaia_cat["dec"])]
    catalog_data = np.array(
        [input_dm.meta.wcs.world_to_pixel(ra, dec) for ra, dec in gaia_source_coords]
    )
    return catalog_data


def create_base_image_source_catalog(
    tmp_path,
    output_filename,
    catalog_data,
    catalog_format: str = "ascii.ecsv",
    save_catalogs=True,
):
    """
    Write a temp CSV file to be used as source catalog, similar to what
    is produced by the previous pipeline step, source detection.

    Parameters
    ----------
    tmp_path : pathlib.PosixPath
        A path-like object representing the path where to save the file.
    output_filename : string
        The output filename (with extension).
    catalog_data : numpy.ndarray
        A numpy array with the (x, y) coordinates of the
        "detected" sources
    catalog_format : str, optional
        A string indicating the catalog format.
    save_catalogs : boolean, optional
        A boolean indicating whether the source catalog should be saved to disk.
    """
    src_detector_coords = catalog_data
    output = os.path.join(tmp_path, output_filename)
    t = table.Table(src_detector_coords, names=("x", "y"))
    if save_catalogs:
        t.write((tmp_path / output), format=catalog_format)
    # mimic the same output format from SourceDetectionStep
    t.add_column([i for i in range(len(t))], name="id", index=0)
    t.add_column([np.float64(i) for i in range(len(t))], name="flux")
    t.rename_columns(["x", "y"], ["xcentroid", "ycentroid"])
    return t.as_array()


def add_tweakreg_catalog_attribute(
    tmp_path,
    input_dm,
    catalog_filename="base_image_sources",
    catalog_data=None,
    catalog_format: str = "ascii.ecsv",
    save_catalogs=True,
):
    """
    Add tweakreg_catalog attribute to the meta, which is a mandatory
    attribute for TweakReg. The tweakreg_catalog attribute contains
    a table with the coordinates of the sources detected by the
    previous step. The coordinates will be used in the process of
    correcting the original WCS.

    Parameters
    ----------
    tmp_path : pathlib.PosixPath
        A path-like object representing the path where to save the file.
    input_dm : roman_datamodels.ImageModel
        A Roman image datamodel.
    catalog_filename : str
        Filename to be used when saving the catalog to disk.
    catalog_data : numpy.ndarray, optional
        A numpy array with the (x, y) coordinates of the
        "detected" sources, by default None (see note below).
    catalog_format : str, optional
        A string indicating the catalog format.
    save_catalogs : boolean, optional
        A boolean indicating whether the source catalog should be saved to disk.

    Note
    ----
    If no catalog_data is provided, a default catalog will be created
    by fetching data from Gaia within a search radius of 100 arcsec
    centered at RA=270, Dec=66.
    """
    tweakreg_catalog_filename = catalog_filename
    if catalog_data is None:
        catalog_data = get_catalog_data(input_dm)

    source_catalog = create_base_image_source_catalog(
        tmp_path,
        tweakreg_catalog_filename,
        catalog_data=catalog_data,
        catalog_format=catalog_format,
        save_catalogs=save_catalogs,
    )

    input_dm.meta["source_detection"] = maker_utils.mk_source_detection()

    if save_catalogs:
        # SourceDetectionStep adds the catalog path+filename
        input_dm.meta.source_detection["tweakreg_catalog_name"] = os.path.join(
            tmp_path, tweakreg_catalog_filename
        )
    else:
        # SourceDetectionStep attaches the catalog data as a structured array
        input_dm.meta.source_detection["tweakreg_catalog"] = source_catalog


@pytest.fixture
def base_image():
    """
    Create a base image with a realistic WCSInfo and a WCS.

    Notes
    -----
    The size of the image needs to be relatively large in order for
    the source catalog step to find a reasonable number of sources in the image.

    shift_1 and shift_2 (units in pixel) are used to shift the WCS projection plane.
    """

    def _base_image(shift_1=0, shift_2=0):
        l2 = maker_utils.mk_level2_image(shape=(2000, 2000))
        l2.meta.exposure.mid_time = Time("2016-01-01T00:00:00")
        # update wcsinfo
        update_wcsinfo(l2)
        # add a dummy WCS object
        create_wcs_for_tweakreg_pipeline(l2, shift_1=shift_1, shift_2=shift_2)
        l2_im = rdm.ImageModel(l2)
        return l2_im

    return _base_image


@pytest.mark.parametrize(
    "input, error_type",
    [
        (list(), (TypeError,)),
        ([""], (TypeError,)),
        (["", ""], (TypeError,)),
        ("", (TypeError,)),
        ([1, 2, 3], (TypeError,)),
    ],
)
def test_tweakreg_raises_error_on_invalid_input(input, error_type):
    # sourcery skip: list-literal
    """Test that TweakReg raises an error when an invalid input is provided."""
    with pytest.raises(Exception) as exec_info:
        trs.TweakRegStep.call(input)

    assert type(exec_info.value) in error_type


def test_tweakreg_raises_attributeerror_on_missing_tweakreg_catalog(base_image):
    """
    Test that TweakReg raises an AttributeError if meta.tweakreg_catalog is missing.
    """
    img = base_image()
    with pytest.raises(Exception) as exec_info:
        trs.TweakRegStep.call([img])

    assert type(exec_info.value) == AttributeError


def test_tweakreg_returns_modelcontainer_on_roman_datamodel_as_input(
    tmp_path, base_image
):
    """Test that TweakReg always returns a ModelContainer when processing an open Roman DataModel as input."""

    img = base_image(shift_1=1000, shift_2=1000)
    add_tweakreg_catalog_attribute(tmp_path, img, catalog_filename="img_1")

    test_input = img

    res = trs.TweakRegStep.call(test_input)
    assert res[0].meta.cal_step.tweakreg == "COMPLETE"
    assert isinstance(res, ModelContainer)


def test_tweakreg_returns_modelcontainer_on_modelcontainer_as_input(
    tmp_path, base_image
):
    """Test that TweakReg always returns a ModelContainer when processing a ModelContainer as input."""

    img = base_image(shift_1=1000, shift_2=1000)
    add_tweakreg_catalog_attribute(tmp_path, img, catalog_filename="img_1")

    test_input = ModelContainer([img])

    res = trs.TweakRegStep.call(test_input)
    assert res[0].meta.cal_step.tweakreg == "COMPLETE"
    assert isinstance(res, ModelContainer)


def test_tweakreg_returns_modelcontainer_on_association_file_as_input(
    tmp_path, base_image
):
    """Test that TweakReg always returns a ModelContainer when processing an association file as input."""

    img_1 = base_image(shift_1=1000, shift_2=1000)
    img_2 = base_image(shift_1=1000, shift_2=1000)
    add_tweakreg_catalog_attribute(tmp_path, img_1, catalog_filename="img_1")
    add_tweakreg_catalog_attribute(tmp_path, img_2, catalog_filename="img_2")
    img_1.save(tmp_path / "img_1.asdf")
    img_2.save(tmp_path / "img_2.asdf")
    asn_filepath = create_asn_file(tmp_path)

    test_input = asn_filepath

    res = trs.TweakRegStep.call(test_input)
    assert all([x.meta.cal_step.tweakreg == "COMPLETE" for x in res])
    assert isinstance(res, ModelContainer)


def test_tweakreg_returns_modelcontainer_on_list_of_asdf_file_as_input(
    tmp_path, base_image
):
    """Test that TweakReg always returns a ModelContainer when processing a list of ASDF files as input."""

    img_1 = base_image(shift_1=1000, shift_2=1000)
    img_2 = base_image(shift_1=1000, shift_2=1000)
    add_tweakreg_catalog_attribute(tmp_path, img_1, catalog_filename="img_1")
    add_tweakreg_catalog_attribute(tmp_path, img_2, catalog_filename="img_2")
    img_1.save(tmp_path / "img_1.asdf")
    img_2.save(tmp_path / "img_2.asdf")

    tmp_path_str = tmp_path.as_posix()
    test_input = [
        f"{tmp_path_str}/img_1.asdf",
        f"{tmp_path_str}/img_2.asdf",
    ]

    res = trs.TweakRegStep.call(test_input)
    assert all([x.meta.cal_step.tweakreg == "COMPLETE" for x in res])
    assert isinstance(res, ModelContainer)


def test_tweakreg_returns_modelcontainer_on_list_of_roman_datamodels_as_input(
    tmp_path, base_image
):
    """Test that TweakReg always returns a ModelContainer when processing a list of open Roman datamodels as input."""
    img_1 = base_image(shift_1=1000, shift_2=1000)
    img_2 = base_image(shift_1=1000, shift_2=1000)
    add_tweakreg_catalog_attribute(tmp_path, img_1, catalog_filename="img_1")
    add_tweakreg_catalog_attribute(tmp_path, img_2, catalog_filename="img_2")
    img_1.save(tmp_path / "img_1.asdf")
    img_2.save(tmp_path / "img_2.asdf")

    test_input = [img_1, img_2]

    res = trs.TweakRegStep.call(test_input)
    assert all([x.meta.cal_step.tweakreg == "COMPLETE" for x in res])
    assert isinstance(res, ModelContainer)


def test_tweakreg_updates_cal_step(tmp_path, base_image):
    """Test that TweakReg updates meta.cal_step with tweakreg = COMPLETE."""
    img = base_image(shift_1=1000, shift_2=1000)
    add_tweakreg_catalog_attribute(tmp_path, img)
    res = trs.TweakRegStep.call([img])

    assert hasattr(res[0].meta.cal_step, "tweakreg")
    assert res[0].meta.cal_step.tweakreg == "COMPLETE"


def test_tweakreg_updates_group_id(tmp_path, base_image):
    """Test that TweakReg updates 'group_id' with a non-zero length string."""
    img = base_image(shift_1=1000, shift_2=1000)
    add_tweakreg_catalog_attribute(tmp_path, img)
    res = trs.TweakRegStep.call([img])

    assert hasattr(res[0].meta, "group_id")
    assert len(res[0].meta.group_id) > 0


@pytest.mark.parametrize(
    "shift_1, shift_2, tolerance, is_small_correction",
    [
        (0, 0, 5, True),
        (0, 3, 5, True),
        (1, 1, 5, True),
        (5, 5, (5**2 + 5**2) ** 0.5, True),
        (5, 5, 5, False),
        (5, 5, 1, False),
        (5, 5, 3, False),
        (5, 10, 5, False),
    ],
)
def test_tweakreg_correction_magnitude(
    shift_1, shift_2, tolerance, is_small_correction, request
):
    """
    Test that TweakReg corrections are within tolerance.
    All the parametrized values are in arcsec.
    """
    img1 = request.getfixturevalue("base_image")(shift_1=1000, shift_2=1000)
    img2 = request.getfixturevalue("base_image")(
        shift_1=1000 + shift_1 / 0.1, shift_2=1000 + shift_2 / 0.1
    )
    img1_wcs = copy.deepcopy(img1.meta.wcs)
    img2_wcs = copy.deepcopy(img2.meta.wcs)

    step = trs.TweakRegStep()
    step.tolerance = tolerance / 10.0

    assert step._is_wcs_correction_small(img1_wcs, img2_wcs) == is_small_correction


@pytest.mark.parametrize(
    "filename_list, expected_common_name",
    (
        (
            [
                "l1-sca1_cal.asdf",
                "l1-sca2_cal.asdf",
                "l1-sca3_cal.asdf",
            ],
            "l1-sca",
        ),
        (
            [
                "l1-270-66-gaia-2016-sca1_cal.asdf",
                "l1-270-66-gaia-2016-sca2_cal.asdf",
                "l1-270-66-gaia-2016-sca3_cal.asdf",
            ],
            "l1-270-66-gaia-2016-sca",
        ),
    ),
)
def test_tweakreg_common_name(filename_list, expected_common_name, request):
    """Test that TweakReg raises an error when an invalid input is provided."""
    img_list = []
    for filename in filename_list:
        img = request.getfixturevalue("base_image")()
        img.meta["filename"] = filename
        img_list.append(img)

    res = trs._common_name(img_list)

    assert res == expected_common_name


@pytest.mark.parametrize(
    "filename_list",
    (
        [
            "l1-sca1_cal.asdf",
            "l1-sca2_cal.asdf",
            "l1-sca3_cal.asdf",
        ],
        [
            "l1-270-66-gaia-2016-sca1_cal.asdf",
            "l1-270-66-gaia-2016-sca2_cal.asdf",
            "l1-270-66-gaia-2016-sca3_cal.asdf",
        ],
    ),
)
def test_tweakreg_common_name_raises_error_on_invalid_input(filename_list):
    """Test that TweakReg raises an error when an invalid input is provided."""
    with pytest.raises(Exception) as exec_info:
        trs._common_name(filename_list)

    assert type(exec_info.value) == TypeError


@pytest.mark.parametrize(
    "abs_refcat",
    (
        "GAIADR1",
        "GAIADR2",
        "GAIADR3",
    ),
)
def test_tweakreg_save_valid_abs_refcat(tmp_path, abs_refcat, request):
    """Test that TweakReg saves the catalog used for absolute astrometry."""
    img = request.getfixturevalue("base_image")(shift_1=1000, shift_2=1000)
    catalog_filename = "ref_catalog.ecsv"
    abs_refcat_filename = f"fit_{abs_refcat.lower()}_ref.ecsv"
    add_tweakreg_catalog_attribute(tmp_path, img, catalog_filename=catalog_filename)

    trs.TweakRegStep.call(
        [img], save_abs_catalog=True, abs_refcat=abs_refcat, catalog_path=str(tmp_path)
    )

    assert os.path.exists(tmp_path / abs_refcat_filename)
    # clean up
    os.remove(tmp_path / abs_refcat_filename)
    os.remove(tmp_path / catalog_filename)


@pytest.mark.parametrize(
    "abs_refcat",
    (None, ""),
)
def test_tweakreg_defaults_to_valid_abs_refcat(tmp_path, abs_refcat, request):
    """Test that TweakReg defaults to DEFAULT_ABS_REFCAT on invalid values."""
    img = request.getfixturevalue("base_image")(shift_1=1000, shift_2=1000)
    catalog_filename = "ref_catalog.ecsv"
    abs_refcat_filename = f"fit_{trs.DEFAULT_ABS_REFCAT.lower()}_ref.ecsv"
    add_tweakreg_catalog_attribute(tmp_path, img, catalog_filename=catalog_filename)

    trs.TweakRegStep.call(
        [img], save_abs_catalog=True, abs_refcat=abs_refcat, catalog_path=str(tmp_path)
    )

    assert os.path.exists(tmp_path / abs_refcat_filename)
    # clean up
    os.remove(tmp_path / abs_refcat_filename)
    os.remove(tmp_path / catalog_filename)


def test_tweakreg_raises_error_on_invalid_abs_refcat(tmp_path, base_image):
    """Test that TweakReg raises an error when an invalid abs_refcat is provided."""
    img = base_image(shift_1=1000, shift_2=1000)
    add_tweakreg_catalog_attribute(tmp_path, img)

    with pytest.raises(Exception) as exec_info:
        trs.TweakRegStep.call([img], save_abs_catalog=True, abs_refcat="my_ref_cat")

    assert type(exec_info.value) == ValueError


def test_tweakreg_combine_custom_catalogs_and_asn_file(tmp_path, base_image):
    """
    Test that TweakRegStep can handle a custom catalog for the members of an ASN file.
    In this case, the user can create a custom catalog file (catfile) for each of the
    members of an ASN file.
    """
    # create custom catalog and input datamodels
    catalog_format = "ascii.ecsv"
    res_dict = create_custom_catalogs(
        tmp_path, base_image, catalog_format=catalog_format
    )
    catfile = res_dict.get("catfile")
    img1, img2, img3 = res_dict.get("datamodels")
    add_tweakreg_catalog_attribute(tmp_path, img1, catalog_filename="img1")
    add_tweakreg_catalog_attribute(tmp_path, img2, catalog_filename="img2")
    add_tweakreg_catalog_attribute(tmp_path, img3, catalog_filename="img3")
    img1.save(tmp_path / "img1.asdf")
    img2.save(tmp_path / "img2.asdf")
    img3.save(tmp_path / "img3.asdf")

    # create ASN file
    asn_filepath = create_asn_file(
        tmp_path,
        members_mapping=[
            {"expname": img1.meta.filename, "exptype": "science"},
            {"expname": img2.meta.filename, "exptype": "science"},
            {"expname": img3.meta.filename, "exptype": "science"},
        ],
    )
    with open(asn_filepath) as f:
        asn_content = json.load(f)

    res = trs.TweakRegStep.call(
        asn_filepath,
        use_custom_catalogs=True,
        catalog_format=catalog_format,
        catfile=catfile,
    )

    assert type(res) == ModelContainer

    assert hasattr(res[0].meta, "asn")

    assert all(
        x.meta.asn["exptype"] == y["exptype"]
        for x, y in zip(res, asn_content["products"][0]["members"])
    )

    assert all(
        x.meta.filename == y.meta.filename for x, y in zip(res, [img1, img2, img3])
    )

    assert all(type(x) == type(y) for x, y in zip(res, [img1, img2, img3]))

    assert all((x.data == y.data).all() for x, y in zip(res, [img1, img2, img3]))


@pytest.mark.parametrize(
    "catalog_format",
    (
        "ascii",
        "ascii.aastex",
        "ascii.basic",
        "ascii.commented_header",
        "ascii.csv",
        "ascii.ecsv",
        "ascii.fixed_width",
        "ascii.fixed_width_two_line",
        "ascii.ipac",
        "ascii.latex",
        "ascii.mrt",
        "ascii.rdb",
        "ascii.rst",
        "ascii.tab",
    ),
)
def test_tweakreg_use_custom_catalogs(tmp_path, catalog_format, base_image):
    """Test that TweakReg can use custom catalogs."""
    # create input datamodels
    res_dict = create_custom_catalogs(
        tmp_path, base_image, catalog_format=catalog_format
    )
    catfile = res_dict.get("catfile")
    img1, img2, img3 = res_dict.get("datamodels")

    trs.TweakRegStep.call(
        [img1, img2, img3],
        use_custom_catalogs=True,
        catalog_format=catalog_format,
        catfile=catfile,
    )

    assert all(img1.meta.tweakreg_catalog) == all(
        table.Table.read(str(tmp_path / "ref_catalog_1"), format=catalog_format)
    )
    assert all(img2.meta.tweakreg_catalog) == all(
        table.Table.read(str(tmp_path / "ref_catalog_2"), format=catalog_format)
    )
    assert all(img3.meta.tweakreg_catalog) == all(
        table.Table.read(str(tmp_path / "ref_catalog_3"), format=catalog_format)
    )


@pytest.mark.parametrize(
    "theta, offset_x, offset_y",
    [
        (0 * u.deg, 0, 0),
        (0 * u.deg, 1, 0),
        (0 * u.deg, 0, 1),
        (0 * u.deg, 1, 1),
        (0.05 * u.deg, 0, 0),
        (0.05 * u.deg, 1, 0),
        (0.05 * u.deg, 0, 1),
        (0.05 * u.deg, 1, 1),
        (0.1 * u.deg, 0, 0),
        (0.1 * u.deg, 1, 0),
        (0.1 * u.deg, 0, 1),
        (0.1 * u.deg, 1, 1),
    ],
)
def test_tweakreg_rotated_plane(tmp_path, theta, offset_x, offset_y, request):
    """
    Test that TweakReg returns accurate results.
    """
    gaia_cat = get_catalog(ra=270, dec=66, sr=100 / 3600)
    gaia_source_coords = [(ra, dec) for ra, dec in zip(gaia_cat["ra"], gaia_cat["dec"])]

    img = request.getfixturevalue("base_image")(shift_1=1000, shift_2=1000)
    original_wcs = copy.deepcopy(img.meta.wcs)

    # calculate original (x,y) for Gaia sources
    original_xy_gaia_sources = np.array(
        [original_wcs.world_to_pixel(ra, dec) for ra, dec in gaia_source_coords]
    )
    # move Gaia sources around by applying linear transformations
    # to their coords in the projected plane (same as a "wrong WCS")
    rot_matrix = models.Rotation2D(angle=theta)
    transformed_xy_gaia_sources = np.array(
        [
            rot_matrix(x, y) + np.array([offset_x, offset_y])
            for x, y in original_xy_gaia_sources
        ]
    )
    # save modified catalog to meta.tweakreg_catalog
    # (coords in the projection plane)
    add_tweakreg_catalog_attribute(
        tmp_path, img, catalog_data=transformed_xy_gaia_sources
    )

    trs.TweakRegStep.call([img], abs_minobj=3)

    # get world coords for Gaia sources using "wrong WCS"
    original_ref_source = [
        original_wcs.pixel_to_world(x, y) for x, y in transformed_xy_gaia_sources
    ]
    # get world coords for Gaia sources using tweaked WCS
    new_ref_source = [
        img.meta.wcs.pixel_to_world(x, y) for x, y in transformed_xy_gaia_sources
    ]
    # celestial coordinates for Gaia sources
    gaia_ref_source = [
        coord.SkyCoord(ra * u.deg, dec * u.deg, frame="icrs")
        for ra, dec in gaia_source_coords
    ]
    # calculate distance between "wrong WCS" result and Gaia
    # (rounded to the 10th decimal place to avoid floating point issues)
    dist1 = [
        np.round(gref.separation(oref), 10)
        for gref, oref in zip(gaia_ref_source, original_ref_source)
    ]
    # calculate distance between tweaked WCS result and Gaia
    # (rounded to the 10th decimal place to avoid floating point issues)
    dist2 = [
        np.round(gref.separation(nref), 10)
        for gref, nref in zip(gaia_ref_source, new_ref_source)
    ]

    assert np.array([np.less_equal(d2, d1) for d1, d2 in zip(dist1, dist2)]).all()


@pytest.mark.parametrize(
    "source_detection_save_catalogs",
    (True, False),
)
def test_remove_tweakreg_catalog_data(
    tmp_path, source_detection_save_catalogs, request
):
    """
    Test to check that the source catalog data is removed from meta
    regardless of selected SourceDetectionStep.save_catalogs option.
    """
    img = request.getfixturevalue("base_image")(shift_1=1000, shift_2=1000)
    add_tweakreg_catalog_attribute(
        tmp_path, img, save_catalogs=source_detection_save_catalogs
    )

    trs.TweakRegStep.call([img])

    assert not hasattr(img.meta.source_detection, "tweakreg_catalog")
    assert hasattr(img.meta, "tweakreg_catalog")


def test_tweakreg_parses_asn_correctly(tmp_path, base_image):
    """Test that TweakReg can parse an ASN file properly."""

    img_1 = base_image(shift_1=1000, shift_2=1000)
    img_2 = base_image(shift_1=1000, shift_2=1000)
    img_1.meta["filename"] = "img_1.asdf"
    img_2.meta["filename"] = "img_2.asdf"
    add_tweakreg_catalog_attribute(tmp_path, img_1, catalog_filename="img_1")
    add_tweakreg_catalog_attribute(tmp_path, img_2, catalog_filename="img_2")
    img_1.save(tmp_path / "img_1.asdf")
    img_2.save(tmp_path / "img_2.asdf")
    asn_filepath = create_asn_file(tmp_path)
    with open(asn_filepath) as f:
        asn_content = json.load(f)

    res = trs.TweakRegStep.call(asn_filepath)
    assert type(res) == ModelContainer

    assert hasattr(res[0].meta, "asn")
    assert (
        res[0].meta.asn["exptype"]
        == asn_content["products"][0]["members"][0]["exptype"]
    )
    assert (
        res[1].meta.asn["exptype"]
        == asn_content["products"][0]["members"][1]["exptype"]
    )
    assert res[0].meta.asn["pool_name"] == asn_content["asn_pool"]
    assert res[1].meta.asn["pool_name"] == asn_content["asn_pool"]

    assert res[0].meta.filename == img_1.meta.filename
    assert res[1].meta.filename == img_2.meta.filename

    assert type(res[0]) == type(img_1)
    assert type(res[1]) == type(img_2)

    assert (res[0].data == img_1.data).all()
    assert (res[1].data == img_2.data).all()


def test_tweakreg_raises_error_on_connection_error_to_the_vo_service(
    tmp_path, base_image, monkeypatch
):
    """
    Test that TweakReg raises an error when there is a connection error with
    the VO API server, which means that an absolute reference catalog cannot be created.
    """

    img = base_image(shift_1=1000, shift_2=1000)
    add_tweakreg_catalog_attribute(tmp_path, img)

    monkeypatch.setattr("requests.get", MockConnectionError)
    res = trs.TweakRegStep.call([img])

    assert type(res) == ModelContainer
    assert len(res) == 1
    assert res[0].meta.cal_step.tweakreg.lower() == "skipped"


def test_fit_results_in_meta(tmp_path, base_image):
    """
    Test that the WCS fit results from tweakwcs are available in the meta tree.
    """

    img = base_image(shift_1=1000, shift_2=1000)
    add_tweakreg_catalog_attribute(tmp_path, img)

    res = trs.TweakRegStep.call([img])

    assert type(res) == ModelContainer
    assert [
        hasattr(x.meta, "wcs_fit_results") and len(x.meta.wcs_fit_results) > 0
        for x in res
    ]


def test_tweakreg_returns_skipped_for_one_file(tmp_path, base_image):
    """
    Test that TweakRegStep assigns meta.cal_step.tweakreg to "SKIPPED"
    when one image is provided but no alignment to a reference catalog is desired.
    """
    img = base_image(shift_1=1000, shift_2=1000)
    add_tweakreg_catalog_attribute(tmp_path, img)

    # disable alignment to absolute reference catalog
    trs.ALIGN_TO_ABS_REFCAT = False
    res = trs.TweakRegStep.call([img])

    assert all(x.meta.cal_step.tweakreg == "SKIPPED" for x in res)


def test_tweakreg_handles_multiple_groups(tmp_path, base_image):
    """
    Test that TweakRegStep can perform relative alignment for all images in the groups
    before performing absolute alignment.
    """
    img1 = base_image(shift_1=1000, shift_2=1000)
    img2 = base_image(shift_1=990, shift_2=990)
    add_tweakreg_catalog_attribute(tmp_path, img1, catalog_filename="img1")
    add_tweakreg_catalog_attribute(tmp_path, img2, catalog_filename="img2")

    img1.meta.observation["program"] = "-program_id1"
    img2.meta.observation["program"] = "-program_id2"

    img1.meta["filename"] = "file1.asdf"
    img2.meta["filename"] = "file2.asdf"

    res = trs.TweakRegStep.call([img1, img2])

    assert len(res.models_grouped) == 2
    all(
        (
            r.meta.group_id.split("-")[1],
            i.meta.observation.program.split("-")[1],
        )
        for r, i in zip(res, [img1, img2])
    )


def test_tweakreg_multiple_groups_valueerror(tmp_path, base_image):
    """
    Test that TweakRegStep throws an error when too few input images or
    groups of images with non-empty catalogs is provided.
    """
    img1 = base_image(shift_1=1000, shift_2=1000)
    img2 = base_image(shift_1=1000, shift_2=1000)
    add_tweakreg_catalog_attribute(tmp_path, img1, catalog_filename="img1")
    add_tweakreg_catalog_attribute(tmp_path, img2, catalog_filename="img2")

    img1.meta.observation["program"] = "-program_id1"
    img2.meta.observation["program"] = "-program_id2"

    trs.ALIGN_TO_ABS_REFCAT = False
    res = trs.TweakRegStep.call([img1, img2])

    assert all(x.meta.cal_step.tweakreg == "SKIPPED" for x in res)


@pytest.mark.parametrize(
    "column_names",
    [("x", "y"), ("xcentroid", "ycentroid")],
)
def test_imodel2wcsim_valid_column_names(tmp_path, base_image, column_names):
    """
    Test that _imodel2wcsim handles different catalog column names.
    """
    img_1 = base_image(shift_1=1000, shift_2=1000)
    img_2 = base_image(shift_1=1030, shift_2=1030)
    add_tweakreg_catalog_attribute(tmp_path, img_1, catalog_filename="img_1")
    add_tweakreg_catalog_attribute(tmp_path, img_2, catalog_filename="img_2")
    # set meta.tweakreg_catalog (this is automatically added by TweakRegStep)
    catalog_format = "ascii.ecsv"
    for x in [img_1, img_2]:
        x.meta["tweakreg_catalog"] = Table.read(
            x.meta.source_detection.tweakreg_catalog_name,
            format=catalog_format,
        )
        x.meta.tweakreg_catalog.rename_columns(("x", "y"), column_names)

    images = ModelContainer([img_1, img_2])
    grp_img = list(images.models_grouped)
    g = grp_img[0]

    step = trs.TweakRegStep()
    imcats = list(map(step._imodel2wcsim, g))

    assert all(x.meta["image_model"]() == y for x, y in zip(imcats, [img_1, img_2]))
    assert np.all(
        x.meta["catalog"] == y.meta.tweakreg_catalog
        for x, y in zip(imcats, [img_1, img_2])
    )


@pytest.mark.parametrize(
    "column_names",
    [
        ("x_centroid", "y_centroid"),
        ("x_cen", "y_cen"),
    ],
)
def test_imodel2wcsim_error_invalid_column_names(tmp_path, base_image, column_names):
    """
    Test that _imodel2wcsim raises a ValueError on invalid catalog column names.
    """
    img_1 = base_image(shift_1=1000, shift_2=1000)
    img_2 = base_image(shift_1=1030, shift_2=1030)
    add_tweakreg_catalog_attribute(tmp_path, img_1, catalog_filename="img_1")
    add_tweakreg_catalog_attribute(tmp_path, img_2, catalog_filename="img_2")
    # set meta.tweakreg_catalog (this is automatically added by TweakRegStep)
    catalog_format = "ascii.ecsv"
    for x in [img_1, img_2]:
        x.meta["tweakreg_catalog"] = Table.read(
            x.meta.source_detection.tweakreg_catalog_name,
            format=catalog_format,
        )
        x.meta.tweakreg_catalog.rename_columns(("x", "y"), column_names)

    images = ModelContainer([img_1, img_2])
    grp_img = list(images.models_grouped)
    g = grp_img[0]

    step = trs.TweakRegStep()
    with pytest.raises(Exception) as exec_info:
        list(map(step._imodel2wcsim, g))

    assert type(exec_info.value) == ValueError


def test_imodel2wcsim_error_invalid_catalog(tmp_path, base_image):
    """
    Test that _imodel2wcsim raises an error on invalid catalog format.
    """
    img_1 = base_image(shift_1=1000, shift_2=1000)
    add_tweakreg_catalog_attribute(tmp_path, img_1, catalog_filename="img_1")
    # set meta.tweakreg_catalog (this is automatically added by TweakRegStep)
    img_1.meta["tweakreg_catalog"] = "nonsense"

    images = ModelContainer([img_1])
    grp_img = list(images.models_grouped)
    g = grp_img[0]

    step = trs.TweakRegStep()
    with pytest.raises(Exception) as exec_info:
        list(map(step._imodel2wcsim, g))

    assert type(exec_info.value) == AttributeError


def test_parse_catfile_valid_catalog(tmp_path, base_image):
    """
    Test that _parse_catfile can parse a custom catalog with valid format.
    """
    # create custom catalog file and input datamodels
    catalog_format = "ascii.ecsv"
    res_dict = create_custom_catalogs(
        tmp_path, base_image, catalog_format=catalog_format
    )
    catfile = res_dict.get("catfile")
    catdict = trs._parse_catfile(catfile)

    assert all(
        x.meta.filename == y for x, y in zip(res_dict.get("datamodels"), catdict.keys())
    )


@pytest.mark.parametrize("catfile", (None, ""))
def test_parse_catfile_returns_none(catfile):
    """
    Test that _parse_catfile returns None when catfile = None or catfile = "".
    """
    catdict = trs._parse_catfile(catfile=catfile)

    assert catdict is None


@pytest.mark.parametrize(
    "catfile_line_content",
    ["img1.asdf\nimg2.asdf\nimg3.asdf"],
)
def test_parse_catfile_returns_none_on_invalid_content(tmp_path, catfile_line_content):
    """
    Test that _parse_catfile returns a dict where all the values are None
    if only filename is present in catfile (i.e. no associated catalog).
    """
    # create custom catalog file and input datamodels
    catfile = str(tmp_path / "catfile.txt")
    catfile_content = StringIO()
    # write empty line to catfile
    catfile_content.write(catfile_line_content)
    # write StringIO object to disk
    with open(catfile, mode="w") as f:
        print(catfile_content.getvalue(), file=f)

    catdict = trs._parse_catfile(catfile)

    assert not all(catdict.values())


@pytest.mark.parametrize(
    "catfile_line_content",
    ["img1.asdf column1 column2 column3"],
)
def test_parse_catfile_raises_error_on_invalid_content(tmp_path, catfile_line_content):
    """
    Test that _parse_catfile raises an error if catfile contains more than
    two columns (i.e. image name and corresponding catalog path).
    """
    # create custom catalog file and input datamodels
    catfile = str(tmp_path / "catfile.txt")
    catfile_content = StringIO()
    # write empty line to catfile
    catfile_content.write(catfile_line_content)
    # write StringIO object to disk
    with open(catfile, mode="w") as f:
        print(catfile_content.getvalue(), file=f)

    with pytest.raises(Exception) as exec_info:
        trs._parse_catfile(catfile)

    assert type(exec_info.value) == ValueError



#! /usr/bin/env python

import roman_datamodels as rdm
from roman_datamodels.datamodels import SaturationRefModel

from romancal.saturation import saturation
from romancal.stpipe import RomanStep

__all__ = ["SaturationStep"]


class SaturationStep(RomanStep):
    """
    This Step sets saturation flags.
    """

    reference_file_types = ["saturation"]

    def process(self, input):
        # Open the input data model
        with rdm.open(input, lazy_load=False) as input_model:
            # Get the name of the saturation reference file
            self.ref_name = self.get_reference_file(input_model, "saturation")

            # Check for a valid reference file
            if self.ref_name == "N/A":
                self.log.warning("No SATURATION reference file found")
                self.log.warning("Saturation step will be skipped")
                result = input_model.copy()
                result.meta.cal_step.saturation = "SKIPPED"
                return result

            # Open the reference file data model
            # Test for reference file
            self.log.info("Using SATURATION reference file: %s", self.ref_name)
            ref_model = SaturationRefModel(self.ref_name)

            # Perform saturation check
            sat = saturation.flag_saturation(input_model, ref_model)

            # Close the reference file and update the step status
            ref_model.close()
            sat.meta.cal_step.saturation = "COMPLETE"

            if self.save_results:
                try:
                    self.suffix = "saturation"
                except AttributeError:
                    self["suffix"] = "saturation"

        return sat



"""Association Registry"""

import importlib.util
import logging
from inspect import getmembers, isclass, isfunction, ismethod, ismodule
from os.path import basename, expanduser, expandvars

from . import libpath
from .exceptions import AssociationError, AssociationNotValidError
from .lib.callback_registry import CallbackRegistry

__all__ = ["AssociationRegistry", "RegistryMarker"]

# Configure logging
logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())

# Library files
_ASN_RULE = "association_rules.py"


class AssociationRegistry(dict):
    """The available associations

    Parameters
    ----------
    definition_files : [str,]
        The files to find the association definitions in.

    include_default : bool
        True to include the default definitions.

    global_constraints : Constraint
        Constraints to be added to each rule.

    name : str
        An identifying string, used to prefix rule names.

    include_bases : bool
        If True, include base classes not considered
        rules.

    Notes
    -----
    The general workflow is as follows:

        * Create the registry
            >>> from romancal.associations.registry import AssociationRegistry
            >>> registry = AssociationRegistry()

        * Create associations from an item
            >>> associations, reprocess = registry.match(item) # doctest: +SKIP

        * Finalize the associations
            >>> final_asns = registry.callback.reduce('finalize', associations) # doctest: +SKIP

    In practice, this is one step in a larger loop over all items to
    be associated. This does not account for adding items to already
    existing associations. See :py:func:`~romancal.associations.generate` for more information.
    """  # noqa: E501

    def __init__(
        self,
        definition_files=None,
        include_default=True,
        global_constraints=None,
        name=None,
        include_bases=False,
    ):
        super().__init__()

        # Generate a UUID for this instance. Used to modify rule
        # names.
        self.name = name

        # Callback registry
        self.callback = CallbackRegistry()

        # Precache the set of rules
        self._rule_set = set()

        if definition_files is None:
            definition_files = []
        if include_default:
            definition_files.insert(0, libpath(_ASN_RULE))
        if len(definition_files) <= 0:
            raise AssociationError("No rule definition files specified.")

        self.schemas = []
        self.Utility = type("Utility", (object,), {})
        for fname in definition_files:
            module = import_from_file(fname)
            self.populate(
                module,
                global_constraints=global_constraints,
                include_bases=include_bases,
            )

    @property
    def rule_set(self):
        """Rules within the Registry"""
        return self._rule_set

    def match(self, item, version_id=None, allow=None, ignore=None):
        """See if item belongs to any of the associations defined.

        Parameters
        ----------
        item : dict
            An item, like from a Pool, to find associations for.

        version_id : str
            If specified, a string appended to association names.
            If None, nothing is used.

        allow : [type(Association), ...]
            List of rules to allow to be matched. If None, all
            available rules will be used.

        ignore : list
            A list of associations to ignore when looking for a match.
            Intended to ensure that already created associations
            are not re-created.

        Returns
        -------
        (associations, reprocess_list): 2-tuple
            associations : [association,...]
                List of associations item belongs to. Empty if none match
            reprocess_list : [AssociationReprocess, ...]
                List of reprocess events.
        """
        if allow is None:
            allow = self.rule_set
        if ignore is None:
            ignore = []
        associations = []
        process_list = []
        for name, rule in self.items():
            if rule not in ignore and rule in allow:
                asn, reprocess = rule.create(item, version_id)
                process_list.extend(reprocess)
                if asn is not None:
                    associations.append(asn)
        return associations, process_list

    def validate(self, association):
        """Validate a given association

        Parameters
        ----------
        association : association-like
            The data to validate

        Returns
        -------
        rules : list
            List of rules that validated

        Raises
        ------
        AssociationNotValidError
            Association did not validate
        """

        # Change rule validation from an exception
        # to a boolean
        def is_valid(rule, association):
            try:
                rule.validate(association)
            except AssociationNotValidError:
                return False
            else:
                return True

        results = [
            rule for rule_name, rule in self.items() if is_valid(rule, association)
        ]

        if len(results) == 0:
            raise AssociationNotValidError(
                f'Structure did not validate: "{association}"'
            )
        return results

    def load(self, serialized, format=None, validate=True, first=True, **kwargs):
        """Load a previously serialized association

        Parameters
        ----------
        serialized : object
            The serialized form of the association.

        format : str or None
            The format to force. If None, try all available.

        validate : bool
            Validate against the class' defined schema, if any.

        first : bool
            A serialization potentially matches many rules.
            Only return the first succesful load.

        kwargs : dict
            Other arguments to pass to the `load` method

        Returns
        -------
        The Association object, or the list of association objects.

        Raises
        ------
        AssociationError
            Cannot create or validate the association.
        """
        results = []
        for rule_name, rule in self.items():
            try:
                results.append(
                    rule.load(serialized, format=format, validate=validate, **kwargs)
                )
            except (AssociationError, AttributeError) as err:
                lasterr = err
                continue
            if first:
                break
        if len(results) == 0:
            raise lasterr
        if first:
            return results[0]
        else:
            return results

    def populate(self, module, global_constraints=None, include_bases=None):
        """Parse out all rules and callbacks in a module and add them to the registry

        Parameters
        ----------
        module : module
            The module, and all submodules, to be parsed.
        """
        for name, obj in get_marked(module, include_bases=include_bases):
            # Add rules.
            # Add rules.
            if (include_bases and isclass(obj)) or obj._asnreg_role == "rule":
                try:
                    self.add_rule(name, obj, global_constraints=global_constraints)
                except TypeError:
                    logger.debug(
                        f"Could not add object {obj} as a rule due to TypeError"
                    )
                continue

            # Add callbacks
            if obj._asnreg_role == "callback":
                for event in obj._asnreg_events:
                    self.callback.add(event, obj)
                    continue

            # Add schema
            if obj._asnreg_role == "schema":
                self.schemas.append(obj._asnreg_schema)
                continue

            # Add utility classes
            if obj._asnreg_role == "utility":
                self.Utility = type("Utility", (obj, self.Utility), {})

    def add_rule(self, name, obj, global_constraints=None):
        """Add object as rule to registry

        Parameters
        ----------
        name : str
            Name of the object

        obj : object
            The object to be considered a rule

        global_constraints : dict
            The global constraints to attach to the rule.
        """
        try:
            rule_name = "_".join([self.name, name])
        except TypeError:
            rule_name = name
        rule = type(rule_name, (obj,), {})
        rule.GLOBAL_CONSTRAINT = global_constraints
        rule.registry = self
        self.__setitem__(rule_name, rule)
        self._rule_set.add(rule)


class RegistryMarker:
    """Mark rules, callbacks, and modules for inclusion into a registry"""

    class Schema:
        def __init__(self, obj):
            self._asnreg_role = "schema"
            self._asnreg_schema = obj
            RegistryMarker.mark(self)

        @property
        def schema(self):
            return self._asnreg_schema

    @staticmethod
    def mark(obj):
        """Mark that an object should be part of the registry

        Parameters
        ----------
        obj : object
            The object to mark

        Returns
        -------
        obj
            Object that has been marked. Returned to enable
            use as a decorator.

        Notes
        -----
        The following attributes are added to the object:

        - _asnreg_mark : True
              Attribute added to object and is set to True

        - _asnreg_role : str or None
              If not already assigned, the role is left
              unspecified using None.

        """
        obj._asnreg_marked = True
        obj._asnreg_role = getattr(obj, "_asnreg_role", None)
        return obj

    @staticmethod
    def rule(obj):
        """Mark object as rule

        Parameters
        ----------
        obj : object
            The object that should be treated as a rule

        Returns
        -------
        obj : object
            Return object to enable use as a decorator.

        Notes
        -----
        The following attributes are added to the object:

        - _asnreg_role : 'rule'
              Attributed added to object and set to `rule`

        - _asnreg_mark : True
              Attributed added to object and set to True
        """
        obj._asnreg_role = "rule"
        RegistryMarker.mark(obj)
        return obj

    @staticmethod
    def callback(event):
        """Mark object as a callback for an event

        Parameters
        ----------
        event : str
            Event this is a callback for.

        obj : func
            Function, or any callable, to be called
            when the corresponding event is triggered.

        Returns
        -------
        func
            Function to use as a decorator for the object to be marked.

        Notes
        -----
        The following attributes are added to the object:

        - _asnreg_role : 'callback'
              The role the object as been assigned.
        - _asnreg_events : [event[, ...]]
              The events this callable object is a callback for.
        - _asnreg_mark : True
              Indicated that the object has been marked.
        """

        def decorator(func):
            try:
                events = func._asnreg_events
            except AttributeError:
                events = list()
            events.append(event)
            RegistryMarker.mark(func)
            func._asnreg_role = "callback"
            func._asnreg_events = events
            return func

        return decorator

    @staticmethod
    def schema(filename):
        """Mark a file as a schema source"""
        schema = RegistryMarker.Schema(filename)
        return schema

    @staticmethod
    def utility(class_obj):
        """Mark the class as a Utility class"""
        class_obj._asnreg_role = "utility"
        RegistryMarker.mark(class_obj)
        return class_obj

    @staticmethod
    def is_marked(obj):
        """Has an objected been marked?"""
        return hasattr(obj, "_asnreg_marked")


# Utilities
def import_from_file(filename):
    """Import a file as a module

    Parameters
    ---------
    filename : str
        The file to import

    Returns
    -------
    module : python module
        The imported module
    """
    path = expandvars(expanduser(filename))
    module_name = basename(path).split(".")[0]
    spec = importlib.util.spec_from_file_location(module_name, path)
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)
    return module


def get_marked(module, predicate=None, include_bases=False):
    """Recursively get all executable objects

    Parameters
    ----------
    module : python module
        The module to examine

    predicate : bool func(object)
        Determinant of what gets returned.
        If None, all object types are examined

    include_bases : bool
        If True, include base classes not considered
        rules.

    Returns
    -------
    class object : generator
        A generator that will yield all class members in the module.
    """

    def is_method(obj):
        return isfunction(obj) or ismethod(obj)

    for name, obj in getmembers(module, predicate):
        if isclass(obj):
            yield from get_marked(obj, predicate=is_method)
            if RegistryMarker.is_marked(obj) or include_bases:
                yield name, obj
        elif RegistryMarker.is_marked(obj):
            if ismodule(obj):
                yield from get_marked(
                    obj, predicate=predicate, include_bases=include_bases
                )
            else:
                yield name, obj



"""
This package provides support for sky background subtraction and equalization
(matching).
"""

import logging

from .skymatch_step import SkyMatchStep

log = logging.getLogger(__name__)
log.setLevel(logging.DEBUG)

__all__ = ["SkyMatchStep"]



import warnings

from roman_datamodels.dqflags import group, pixel

__all__ = ["pixel", "group"]

warnings.warn(
    "romancal.dqflags is deprecated. Please use roman_datamodels.dqflags instead.",
    DeprecationWarning,
    stacklevel=2,
)



from pathlib import Path
from typing import Optional

import asdf
import numpy


def command():
    try:
        import typer
        from stpreview.downsample import downsample_asdf_to
        from stpreview.image import (
            north_pole_angle,
            percentile_normalization,
            write_image,
        )
        from typing_extensions import Annotated
    except (ImportError, ModuleNotFoundError):
        raise ImportError(
            'SDP requirements not installed; do `pip install "romancal[sdp]"`'
        )

    app = typer.Typer()

    @app.command()
    def preview(
        input: Annotated[
            Path, typer.Argument(help="path to ASDF file with 2D image data")
        ],
        output: Annotated[
            Optional[Path], typer.Argument(help="path to output image file")
        ] = None,
        shape: Annotated[
            Optional[tuple[int, int]],
            typer.Argument(help="desired pixel resolution of output image"),
        ] = (1080, 1080),
        compass: Annotated[
            Optional[bool],
            typer.Option(help="whether to draw a north arrow on the image"),
        ] = True,
    ):
        """
        create a preview image with a north arrow overlay indicating orientation
        """

        if output is None:
            output = Path.cwd()
        if output.is_dir():
            output = output / f"{input.stem}.png"

        with asdf.open(input) as file:
            model = file["roman"]["meta"]["model_type"]
            if "image" not in model.lower():
                raise NotImplementedError(f'"{model}" model not supported')
            wcs = file["roman"]["meta"]["wcs"]

        data = downsample_asdf_to(input=input, shape=shape, func=numpy.nanmean)

        write_image(
            data,
            output,
            shape=shape,
            normalization=percentile_normalization(data, percentile=90),
            colormap="afmhot",
            north_arrow_angle=north_pole_angle(wcs).degree - 90,
        )

    @app.command()
    def thumbnail(
        input: Annotated[
            Path, typer.Argument(help="path to ASDF file with 2D image data")
        ],
        output: Annotated[
            Optional[Path], typer.Argument(help="path to output image file")
        ] = None,
        shape: Annotated[
            Optional[tuple[int, int]],
            typer.Argument(help="desired pixel resolution of output image"),
        ] = (300, 300),
        compass: Annotated[
            Optional[bool],
            typer.Option(help="whether to draw a north arrow on the image"),
        ] = False,
    ):
        if output is None:
            output = Path.cwd()
        if output.is_dir():
            output = output / f"{input.stem}_thumb.png"

        with asdf.open(input) as file:
            model = file["roman"]["meta"]["model_type"]
            if "image" not in model.lower():
                raise NotImplementedError(f'"{model}" model not supported')

        data = downsample_asdf_to(input=input, shape=shape, func=numpy.nanmean)

        write_image(
            data,
            output,
            shape=shape,
            normalization=percentile_normalization(data, percentile=90),
            colormap="afmhot",
        )

    app()


if __name__ == "__main__":
    command()



#!/usr/bin/env python
import logging
import re
from os.path import basename, isfile

import asdf
import numpy as np
from astropy import coordinates
from astropy import units as u
from astropy.modeling import models
from gwcs import WCS, coordinate_frames

import romancal.datamodels.filetype as filetype
from romancal.datamodels import ModelContainer

# step imports
from romancal.flux import FluxStep
from romancal.outlier_detection import OutlierDetectionStep
from romancal.patch_match import patch_match
from romancal.resample import ResampleStep
from romancal.skymatch import SkyMatchStep

from ..stpipe import RomanPipeline

__all__ = ["HighLevelPipeline"]

# Define logging
log = logging.getLogger()
log.setLevel(logging.DEBUG)


class HighLevelPipeline(RomanPipeline):
    """
    HighLevelPipeline: Apply all calibration steps to the roman data
    to produce level 3 products. Included steps are:
    ``skymatch``, ``outlier_detection`` and ``resample``.
    """

    class_alias = "roman_hlp"
    spec = """
        save_results = boolean(default=False)
    """

    # Define aliases to steps
    step_defs = {
        "flux": FluxStep,
        "skymatch": SkyMatchStep,
        "outlier_detection": OutlierDetectionStep,
        "resample": ResampleStep,
    }

    # start the actual processing
    def process(self, input):
        """Process the Roman WFI data from Level 2 to Level 3"""

        log.info("Starting Roman high level calibration pipeline ...")
        if isinstance(input, str):
            input_filename = basename(input)
        else:
            input_filename = None

        # open the input file
        file_type = filetype.check(input)
        if file_type == "asdf":
            log.info("The level three pipeline input needs to be an association")
            return

        if file_type == "asn":
            input = ModelContainer(input)
            self.flux.suffix = "flux"
            result = self.flux(input)
            self.skymatch.suffix = "skymatch"
            result = self.skymatch(result)
            self.outlier_detection.suffix = "outlier_detection"
            result = self.outlier_detection(result)
            #
            # check to see if the product name contains a skycell name & if true get the skycell record
            product_name = input.asn_table["products"][0]["name"]
            try:
                skycell_name = product_name.split("_")[3]
            except IndexError:
                skycell_name = ""
            skycell_record = []

            # if this is a valid skycell name load the database and get the skycell record
            if re.match(r"r\d{3}\w{2}\d{2}x\d{2}y\d{2}", skycell_name):
                skycell_record = patch_match.PATCH_TABLE[
                    np.where(patch_match.PATCH_TABLE["name"][:] == skycell_name)[0][0]
                ]
                log.info("Skycell record %s:", skycell_record)

                if skycell_name in skycell_record["name"]:
                    # skycell name are in the form of r270dm90x99y99
                    # example of product name "r0099101001001001001_F158_visit_r270dm90x99y99"
                    skycell_root = re.findall("^[^%\n\r]*_", product_name)[0]
                    skycell_file_name = skycell_root + skycell_name + "_i2d.asdf"

                    # check to see if there exists a skycell on disk if not create it
                    if not isfile(skycell_file_name):
                        # extract the wcs info from the record for generate_tan_wcs
                        log.info(
                            "Creating skycell image at ra: %f  dec %f",
                            float(skycell_record["ra_center"]),
                            float(skycell_record["dec_center"]),
                        )
                        skycell_wcs = generate_tan_wcs(skycell_record)
                        # skycell_wcs.bounding_box = bounding_box

                        # For resample to use an external grid we need to pass it the skycell gwcs object
                        # Currently we cannot do that directly so create an asdf file to read the skycell gwcs object
                        wcs_tree = {"wcs": skycell_wcs}
                        wcs_file = asdf.AsdfFile(wcs_tree)
                        wcs_file.write_to("skycell_wcs.asdf")
                        self.resample.output_wcs = "skycell_wcs.asdf"
                        self.resample.output_shape = (
                            int(skycell_record["nx"]),
                            int(skycell_record["ny"]),
                        )
                        log.info(
                            "Resampling using %s  and data shape %s",
                            self.resample.output_wcs,
                            self.resample.output_shape,
                        )
                        wcs_file = asdf.open(self.resample.output_wcs)
                        self.suffix = "i2d"
                        result = self.resample(result)
                        self.output_file = input.asn_table["products"][0]["name"]
                    else:
                        log.info("resampling a mosaic file is not yet supported")
                        exit(0)

            else:
                self.resample.suffix = "i2d"
                result = self.resample(result)
                self.suffix = "i2d"
                if input_filename:
                    result.meta.filename = self.output_file
                self.output_file = input.asn_table["products"][0]["name"]

        return result


def generate_tan_wcs(skycell_record, shiftx=0, shifty=0):
    # extract the wcs info from the record for generate_tan_wcs
    # we need the scale, ra, dec, bounding_box

    scale = float(skycell_record["pixel_scale"])
    ra_center = float(skycell_record["ra_center"])
    dec_center = float(skycell_record["dec_center"])
    bounding_box = (
        (-0.5, float(skycell_record["x_center"]) + 0.5),
        (-0.5, float(skycell_record["y_center"]) + 0.5),
    )
    shiftx = bounding_box[0][1]
    shifty = bounding_box[1][1]

    # components of the model
    # shift = models.Shift(shiftx) & models.Shift(shifty)

    # select a scale for the skycell image, this will come from INS and may
    # be optimized for the different survey programs
    scale_x = scale
    scale_y = scale
    # set the pixelsscale to 0.1 arcsec/pixel
    pixelscale = models.Scale(scale_x / 3600.0) & models.Scale(scale_y / 3600.0)

    pixelshift = models.Shift(-1.0 * shiftx) & models.Shift(-1.0 * shifty)
    tangent_projection = models.Pix2Sky_TAN()
    celestial_rotation = models.RotateNative2Celestial(ra_center, dec_center, 180.0)
    det2sky = pixelshift | pixelscale | tangent_projection | celestial_rotation
    detector_frame = coordinate_frames.Frame2D(
        name="detector", axes_names=("x", "y"), unit=(u.pix, u.pix)
    )
    sky_frame = coordinate_frames.CelestialFrame(
        reference_frame=coordinates.ICRS(), name="icrs", unit=(u.deg, u.deg)
    )
    wcsobj = WCS([(detector_frame, det2sky), (sky_frame, None)])
    wcsobj.bounding_box = bounding_box

    return wcsobj



"""Test utility update_path"""

from romancal.associations.asn_from_list import asn_from_list
from romancal.associations.lib.rules_elpp_base import DMS_ELPP_Base
from romancal.associations.lib.update_path import update_path


def test_update_path_level2():
    members = ["a", "b", "c"]
    new_path = "new_path"
    product_name = "new_product"
    asn = asn_from_list(members, product_name=product_name, rule=DMS_ELPP_Base)
    update_path(asn, new_path)
    for product in asn["products"]:
        for member in product["members"]:
            assert member["expname"].startswith(new_path)


def test_update_path_level3():
    members = ["a", "b", "c"]
    new_path = "new_path"
    asn = asn_from_list(members, product_name="test")
    update_path(asn, new_path)
    for product in asn["products"]:
        for member in product["members"]:
            assert member["expname"].startswith(new_path)



#! /usr/bin/env python

import numpy as np
import roman_datamodels as rdm
from roman_datamodels import maker_utils
from roman_datamodels.datamodels import RampModel
from roman_datamodels.dqflags import pixel

from romancal.dq_init import dq_initialization
from romancal.stpipe import RomanStep

__all__ = ["DQInitStep"]


class DQInitStep(RomanStep):
    """Initialize the Data Quality extension from the
    mask reference file.

    The dq_init step initializes the pixeldq attribute of the
    input datamodel using the MASK reference file.  For some
    Guiding and Image model types, initalize the dq attribute of
    the input model instead.  The dq attribute of the MASK model
    is bitwise OR'd with the pixeldq (or dq) attribute of the
    input model.
    """

    reference_file_types = ["mask"]

    def process(self, input):
        """Perform the dq_init calibration step

        Parameters
        ----------
        input : Roman datamodel
            input roman datamodel

        Returns
        -------
        output_model : Roman datamodel
            result roman datamodel
        """
        # Open datamodel
        input_model = rdm.open(input, lazy_load=False)
        # Convert to RampModel if needed
        if not isinstance(input_model, RampModel):
            # Create base ramp node with dummy values (for validation)
            input_ramp = maker_utils.mk_ramp(shape=input_model.shape)
            # check if the input model has a resultantdq from SDF
            if hasattr(input_model, "resultantdq"):
                input_ramp.groupdq = input_model.resultantdq.copy()

            # Copy input_model contents into RampModel
            for key in input_model.keys():
                # check for resultantdq if present copy this to the emp
                # it to the ramp model, we don't want to carry this around
                if key != "resultantdq":
                    if key not in input_ramp:
                        input_ramp[key] = input_model.__getattr__(key)
                    elif isinstance(input_ramp[key], dict):
                        # If a dictionary (like meta), overwrite entires (but keep
                        # required dummy entries that may not be in input_model)
                        input_ramp[key].update(input_model.__getattr__(key))
                    elif isinstance(input_ramp[key], np.ndarray):
                        # Cast input ndarray as RampModel dtype
                        input_ramp[key] = input_model.__getattr__(key).astype(
                            input_ramp[key].dtype
                        )
                    else:
                        input_ramp[key] = input_model.__getattr__(key)

            # Create model from node
            output_model = RampModel(input_ramp)
        else:
            output_model = input_model

        # guide window range information
        x_start = input_model.meta.guidestar.gw_window_xstart
        x_end = input_model.meta.guidestar.gw_window_xsize + x_start
        # set pixeldq array to GW_AFFECTED_DATA (2**4) for the given range
        output_model.pixeldq[int(x_start) : int(x_end), :] = pixel.GW_AFFECTED_DATA
        self.log.info(
            f"Flagging rows from: {x_start} to {x_end} as affected by guide window read"
        )

        # Get reference file path
        reference_file_name = self.get_reference_file(output_model, "mask")

        # Test for reference file
        if reference_file_name != "N/A" and reference_file_name is not None:
            # If there are mask files, perform dq step
            # Open the relevant reference files as datamodels
            reference_file_model = rdm.open(reference_file_name)
            self.log.debug(f"Using MASK ref file: {reference_file_name}")

            # Apply the DQ step, in place
            dq_initialization.do_dqinit(
                output_model,
                reference_file_model,
            )

            # copy original border reference file arrays (data and dq)
            # to their own attributes. they will also remain attached to
            # the science data until they are trimmed at ramp_fit
            # these arrays include the overlap regions in the corners

            output_model.border_ref_pix_right = output_model.data[:, :, -4:].copy()
            output_model.border_ref_pix_left = output_model.data[:, :, :4].copy()
            output_model.border_ref_pix_top = output_model.data[:, :4, :].copy()
            output_model.border_ref_pix_bottom = output_model.data[:, -4:, :].copy()

            output_model.dq_border_ref_pix_right = output_model.pixeldq[:, -4:].copy()
            output_model.dq_border_ref_pix_left = output_model.pixeldq[:, :4].copy()
            output_model.dq_border_ref_pix_top = output_model.pixeldq[:4, :].copy()
            output_model.dq_border_ref_pix_bottom = output_model.pixeldq[-4:, :].copy()

        else:
            # Skip DQ step if no mask files
            reference_file_model = None
            self.log.warning("No MASK reference file found.")
            self.log.warning("DQ initialization step will be skipped.")

            output_model.meta.cal_step.dq_init = "SKIPPED"

        # Close the input and reference files
        input_model.close()

        try:
            reference_file_model.close()
        except AttributeError:
            pass

        if self.save_results:
            try:
                self.suffix = "dqinit"
            except AttributeError:
                self["suffix"] = "dqinit"

        return output_model



"""
Roman Calibration Pipeline base class
"""

import importlib.metadata
import logging
import time

import roman_datamodels as rdm
from roman_datamodels.datamodels import ImageModel, MosaicModel
from stpipe import Pipeline, Step, crds_client

from ..lib.suffix import remove_suffix

_LOG_FORMATTER = logging.Formatter(
    "%(asctime)s.%(msecs)03dZ :: %(name)s :: %(levelname)s :: %(message)s",
    datefmt="%Y-%m-%dT%H:%M:%S",
)
_LOG_FORMATTER.converter = time.gmtime


log = logging.getLogger(__name__)
log.setLevel(logging.DEBUG)


class RomanStep(Step):
    """
    Base class for Roman calibration pipeline steps.
    """

    spec = """
    output_ext =  string(default='.asdf')    # Default type of output
    """

    @classmethod
    def _datamodels_open(cls, init, **kwargs):
        """
        Provide access to this package's datamodels.open function
        so that the stpipe infrastructure knows how to instantiate
        models.
        """
        return rdm.open(init, **kwargs)

    def finalize_result(self, model, reference_files_used):
        """
        Hook that allows the Step to set metadata on the output model
        before save.

        Parameters
        ----------
        model : roman_datamodels.datamodels.DataModel
            Output model.

        reference_files_used : list of tuple(str, str)
            List of reference files used.  The first element of each tuple
            is the reftype code, the second element is the filename.
        """

        model.meta.calibration_software_version = importlib.metadata.version("romancal")

        if isinstance(model, (ImageModel, MosaicModel)):
            for log_record in self.log_records:
                model.cal_logs.append(_LOG_FORMATTER.format(log_record))

        if len(reference_files_used) > 0:
            for ref_name, ref_file in reference_files_used:
                if hasattr(model.meta.ref_file, ref_name):
                    setattr(model.meta.ref_file, ref_name, ref_file)
                    # getattr(model.meta.ref_file, ref_name).name = ref_file
            model.meta.ref_file.crds.sw_version = crds_client.get_svn_version()
            model.meta.ref_file.crds.context_used = crds_client.get_context_used(
                model.crds_observatory
            )

            # this will only run if 'parent' is none, which happens when an individual
            # step is being run or if self is a RomanPipeline and not a RomanStep.
            if self.parent is None:
                log.info(
                    f"Results used CRDS context: {model.meta.ref_file.crds.context_used}"
                )

    def record_step_status(self, model, step_name, success=True):
        """
        Record step completion status in the model's metadata.

        Parameters
        ----------
        model : roman_datamodels.datamodels.DataModel
            Output model.
        step_name : str
            Calibration step name.
        success : bool
            If True, then the step was run successfully.
        """
        # JWST sets model.meta.cal_step.<step name> here.  Roman
        # may do the same, depending on how the metadata format
        # turns out.  Seems like we might be able to combine this
        # with finalize_result somehow.
        pass

    def remove_suffix(self, name):
        """
        Remove any Roman step-specific suffix from the given filename.

        Parameters
        ----------
        name : str
            Filename.

        Returns
        -------
        str
            Filename with step suffix removed.
        """
        # JWST maintains a list of relevant suffixes that is monitored
        # by tests to be up-to-date.  Roman will likely need to do
        # something similar.
        return remove_suffix(name)


# RomanPipeline needs to inherit from Pipeline, but also
# be a subclass of RomanStep so that it will pass checks
# when constructing a pipeline using RomanStep class methods.
class RomanPipeline(Pipeline, RomanStep):
    pass



"""
Unit tests for dark current correction
"""

import numpy as np
import pytest
import roman_datamodels as rdm
from astropy import units as u
from roman_datamodels import maker_utils
from roman_datamodels.datamodels import DarkRefModel, RampModel

from romancal.dark_current import DarkCurrentStep


@pytest.mark.parametrize(
    "instrument, exptype",
    [
        ("WFI", "WFI_IMAGE"),
    ],
)
def test_dark_step_interface(instrument, exptype):
    """Test that the basic inferface works for data requiring a DARK reffile"""

    # Set test size
    shape = (2, 20, 20)

    # Create test ramp and dark models
    ramp_model, darkref_model = create_ramp_and_dark(shape, instrument, exptype)

    # Perform Dark Current subtraction step
    result = DarkCurrentStep.call(ramp_model, override_dark=darkref_model)

    # Test dark results
    assert (result.data == ramp_model.data).all()
    assert type(result) == RampModel
    assert result.validate() is None
    assert result.data.shape == shape
    assert result.groupdq.shape == shape
    assert result.pixeldq.shape == shape[1:]
    assert result.meta.cal_step.dark == "COMPLETE"
    assert result.data.dtype == np.float32
    assert result.pixeldq.dtype == np.uint32
    assert result.groupdq.dtype == np.uint8


@pytest.mark.parametrize(
    "instrument, exptype",
    [
        ("WFI", "WFI_IMAGE"),
    ],
)
def test_dark_step_subtraction(instrument, exptype):
    """Test that the values in a dark reference file are properly subtracted"""

    # Set test size
    shape = (2, 20, 20)

    # Create test ramp and dark models
    ramp_model, darkref_model = create_ramp_and_dark(shape, instrument, exptype)

    # populate data array of science cube
    for i in range(0, 20):
        ramp_model.data[0, 0, i] = i * ramp_model.data.unit
        darkref_model.data[0, 0, i] = i * 0.1 * darkref_model.data.unit
    orig_model = ramp_model.copy()

    # Perform Dark Current subtraction step
    result = DarkCurrentStep.call(ramp_model, override_dark=darkref_model)

    # check that the dark file is subtracted frame by frame from the science data
    diff = orig_model.data.value - darkref_model.data.value

    # test that the output data file is equal to the difference found when subtracting
    # reffile from sci file
    np.testing.assert_array_equal(
        result.data.value, diff, err_msg="dark file should be subtracted from sci file "
    )


@pytest.mark.parametrize(
    "instrument, exptype",
    [
        ("WFI", "WFI_IMAGE"),
    ],
)
def test_dark_step_output_dark_file(tmp_path, instrument, exptype):
    """Test that the the step can output a proper (optional) dark file"""
    path = str(tmp_path / "dark_out.asdf")

    # Set test size
    shape = (2, 20, 20)

    # Create test ramp and dark models
    ramp_model, darkref_model = create_ramp_and_dark(shape, instrument, exptype)

    # Perform Dark Current subtraction step
    DarkCurrentStep.call(ramp_model, override_dark=darkref_model, dark_output=path)

    # Open dark file
    dark_out_file_model = rdm.open(path)

    # Test dark file results
    assert type(dark_out_file_model) == DarkRefModel
    assert dark_out_file_model.validate() is None
    assert dark_out_file_model.data.shape == shape
    assert dark_out_file_model.dq.shape == shape[1:]


@pytest.mark.parametrize(
    "instrument, exptype",
    [
        ("WFI", "WFI_IMAGE"),
    ],
)
def test_dark_step_getbestrefs(tmp_path, instrument, exptype):
    """Test that the the step will skip if CRDS returns N/A for the ref file"""
    path = str(tmp_path / "dark_out.asdf")

    # Set test size
    shape = (2, 20, 20)

    # Create test ramp and dark models
    ramp_model, darkref_model = create_ramp_and_dark(shape, instrument, exptype)

    # Perform Dark Current subtraction step with override = N/A
    result = DarkCurrentStep.call(ramp_model, override_dark="N/A")
    assert result.meta.cal_step.dark == "SKIPPED"

    # Perform Dark Current subtraction step
    DarkCurrentStep.call(ramp_model, override_dark=darkref_model, dark_output=path)

    # Open dark file
    dark_out_file_model = rdm.open(path)

    # Test dark file results
    assert type(dark_out_file_model) == DarkRefModel
    assert dark_out_file_model.validate() is None
    assert dark_out_file_model.data.shape == shape
    assert dark_out_file_model.dq.shape == shape[1:]


def create_ramp_and_dark(shape, instrument, exptype):
    """Helper function to create test ramp and dark models"""

    # Create test ramp model
    ramp = maker_utils.mk_ramp(shape=shape)
    ramp.meta.instrument.name = instrument
    ramp.meta.instrument.detector = "WFI01"
    ramp.meta.instrument.optical_element = "F158"
    ramp.meta.exposure.type = exptype
    ramp.data = u.Quantity(np.ones(shape, dtype=np.float32), u.DN, dtype=np.float32)
    ramp_model = RampModel(ramp)

    # Create dark model
    darkref = maker_utils.mk_dark(shape=shape)
    darkref_model = DarkRefModel(darkref)

    return ramp_model, darkref_model



""" Roman tests for source catalog creation """

import asdf
import pytest
from metrics_logger.decorators import metrics_logger

from romancal.source_catalog.source_catalog_step import SourceCatalogStep
from romancal.stpipe import RomanStep


def passfail(bool_expr):
    """set pass fail"""
    if bool_expr:
        return "Pass"
    return "Fail"


def check_catalog_fields(model, log, modeltype):
    fields = model.dtype.names
    has_pos = ("ra_centroid" in fields) and ("dec_centroid" in fields)
    log.info(
        f"DMS374 MSG: {modeltype} Catalog contains sky coordinates of sources? :"
        + passfail(has_pos)
    )

    has_flux = "aper_total_flux" in fields
    log.info(f"DMS375 MSG: {modeltype} Catalog contains fluxes? :" + passfail(has_flux))

    has_type = "is_extended" in fields
    log.info(
        f"DMS376 MSG: {modeltype} Catalog contains source classification? :"
        + passfail(has_type)
    )

    has_flux_err = "aper_total_flux_err" in fields
    log.info(
        f"DMS386 MSG: {modeltype} Catalog contains flux uncertainties? :"
        + passfail(has_flux_err)
    )

    has_flags = "flags" in fields
    log.info(
        f"DMS387 MSG: {modeltype} Catalog contains DQ flags? :" + passfail(has_flags)
    )

    return has_pos & has_flux & has_type & has_flux_err & has_flags


@pytest.mark.bigdata
@pytest.mark.soctests
@metrics_logger("DMS374", "DMS375", "DMS376", "DMS386", "DMS387")
def test_catalog_l3(rtdata, ignore_asdf_paths):
    # DMS374: positions on ICRF
    # DMS375: fluxes
    # DMS376: type of source
    # DMS386: flux uncertainties
    # DMS387: DQ flags
    inputfn = "r0099101001001001001_F158_visit_i2d.asdf"
    outputfn = inputfn.replace("_i2d", "_cat")
    rtdata.get_data(f"WFI/image/{inputfn}")
    rtdata.input = inputfn
    rtdata.get_truth(f"truth/WFI/image/{outputfn}")

    args = [
        "romancal.step.SourceCatalogStep",
        rtdata.input,
    ]
    RomanStep.from_cmdline(args)
    catalogfp = asdf.open(outputfn)
    catalog = catalogfp["roman"]["source_catalog"]
    step = SourceCatalogStep()
    assert check_catalog_fields(catalog, step.log, "L3")

    # no compare_asdf on the catalogs


@pytest.mark.bigdata
@pytest.mark.soctests
@metrics_logger("DMS374", "DMS375", "DMS376", "DMS386", "DMS387")
def test_catalog_l2(rtdata, ignore_asdf_paths):
    # DMS374: positions on ICRF
    # DMS375: fluxes
    # DMS376: type of source
    # DMS386: flux uncertainties
    # DMS387: DQ flags
    inputfn = "r0000101001001001001_01101_0001_WFI01_cal.asdf"
    outputfn = inputfn.replace("cal", "cat")
    rtdata.get_data(f"WFI/image/{inputfn}")
    rtdata.input = inputfn
    rtdata.get_truth(f"truth/WFI/image/{outputfn}")

    args = [
        "romancal.step.SourceCatalogStep",
        rtdata.input,
    ]
    RomanStep.from_cmdline(args)
    catalogfp = asdf.open(outputfn)
    catalog = catalogfp["roman"]["source_catalog"]
    step = SourceCatalogStep()

    assert check_catalog_fields(catalog, step.log, "L2")
    # no compare_asdf on the catalogs



"""
Create a source catalog for tweakreg
"""

import logging

import astropy.units as u
import numpy as np
from asdf import AsdfFile
from astropy.stats import SigmaClip
from astropy.table import Table
from photutils.background import (
    Background2D,
    MeanBackground,
    MedianBackground,
    ModeEstimatorBackground,
)
from photutils.detection import DAOStarFinder
from roman_datamodels import datamodels as rdm
from roman_datamodels import maker_utils
from roman_datamodels.dqflags import pixel

from romancal.lib import psf
from romancal.stpipe import RomanStep

log = logging.getLogger(__name__)
log.setLevel(logging.DEBUG)

__all__ = ["SourceDetectionStep"]


class SourceDetectionStep(RomanStep):
    """
    SourceDetectionStep: Detect point-like sources in image to create a catalog
    for alignment in tweakreg.
    """

    spec = """
        kernel_fwhm = float(default=None)  # DAOStarFinder:Size of Gaussian kernel,
        # in pixels.
        sharplo = float(default=0.)  # DAOStarFinder: Lower bound for sharpness.
        # Typical values of sharpness range from 0 (flat) to 1 (delta function).
        sharphi = float(default=1.0)  # DAOStarFinder: Upper bound for sharpness.
        roundlo = float(default=-1.0)  # DAOStarFinder: Lower bound for roundness.
        # A circular source will have a zero roundness. A source extended in x or
        # y will have a negative or positive roundness, respectively.
        roundhi = float(default=1.0)  # DAOStarFinder: Upper bound for roundness.
        peakmax = float(default=100000.0)  # Upper limit on brightest pixel in sources.
        max_sources = float(default=None)  # Max number of sources, choosing brightest.
        scalar_threshold = float(default=None) # Detection threshold, to
        # be used for entire image. Assumed to be in same units as data, and is
        # an absolute threshold over background.
        calc_threshold = boolean(default=True) # Calculate a single absolute
        # detection threshold from image based on background.
        snr_threshold = float(default=3.0)  # if calc_threshold_img,
        # the SNR for the threshold image.
        bkg_estimator = string(default='median')  # if calc_threshold_img,
        # choice of mean, median, or mode.
        bkg_boxsize = integer(default=9)  # if calc_threshold_img,
        # size of box in pixels for 2D background.
        bkg_sigma = float(default=2.0) # if calc_threshold_img,
        # n sigma for sigma clipping bkgrnd.
        bkg_filter_size = integer(default=3) # if calc_threshold_img,
        # size of Gaussian kernel for background.
        save_catalogs = boolean(default=False) # Save source catalog to file?
        # Will overwrite an existing catalog of the same name.
        output_cat_filetype = option('asdf', 'ecsv', default='asdf') # Used if
        #save_catalogs=True - file type of output catalog.
        fit_psf = boolean(default=True)  # fit source PSFs for accurate astrometry
        min_separation = integer(default=11)  # don't find multiple sources closer
        # than this number of pixels
    """

    def process(self, input):
        with rdm.open(input, lazy_load=False) as input_model:
            if input_model.meta.exposure.type != "WFI_IMAGE":
                # Check to see if attempt to find sources in non-Image data
                log.info("Skipping source detection for spectral exposure.")
                input_model.meta.cal_step.source_detection = "SKIPPED"
                return input_model

            # remove units from data in this step.
            # DAOStarFinder requires unitless input
            if hasattr(input_model.data, "unit"):
                self.data = input_model.data.value
            else:
                self.data = input_model.data

            # mask DO_NOT_USE pixels

            self.coverage_mask = ((pixel.DO_NOT_USE) & input_model.dq).astype(bool)

            filt = input_model.meta.instrument["optical_element"]

            # if a pre-determined threshold value for detection for the whole
            # image is provided, use this
            if self.scalar_threshold is not None:
                threshold = float(self.scalar_threshold)
                log.info(f"Using a detection threshold of {threshold}.")

            # otherwise, if specified, calculate a scalar threshold from the
            # image by calculating a 2D background image, using this to create
            # a 2d threshold image, and using the median of the 2d threshold
            # image as the scalar detection threshold for the whole image
            elif self.calc_threshold:
                log.info("Determining detection threshold from image.")
                bkg = self._calc_2D_background()
                threshold_img = bkg.background + self.snr_threshold * bkg.background_rms
                threshold = np.median(threshold_img)
                log.info(f"Calculated a detection threshold of {threshold} from image.")
            else:
                threshold = np.median(self.data)

            if self.kernel_fwhm is None:
                # estimate the FWHM of the PSF using rough estimate
                # from the diffraction limit:
                central_wavelength = psf.filter_central_wavelengths[filt] * u.um
                diffraction_limit = float(1.22 * central_wavelength / (2.4 * u.m))
                assume_pixel_scale = 0.11 * u.arcsec / u.pix
                expect_psf_pix = np.sqrt(
                    ((diffraction_limit * u.rad).to(u.arcsec) / assume_pixel_scale) ** 2
                    + 1 * u.pix**2
                ).to_value(u.pix)
                self.kernel_fwhm = expect_psf_pix

            log.info("Detecting sources with DAOStarFinder, using entire image array.")
            daofind = DAOStarFinder(
                fwhm=self.kernel_fwhm,
                threshold=threshold,
                sharplo=self.sharplo,
                sharphi=self.sharphi,
                roundlo=self.roundlo,
                roundhi=self.roundhi,
                brightest=self.max_sources,
                peakmax=self.peakmax,
                min_separation=self.min_separation,
            )

            if self.scalar_threshold is not None:
                # if absolute threshold is provided
                sources = daofind(self.data, mask=self.coverage_mask)

            elif self.calc_threshold is not None:
                # subtract background from data if calculating abs. threshold
                sources = daofind(self.data - bkg.background, mask=self.coverage_mask)
            else:
                sources = daofind(self.data, mask=self.coverage_mask)

            # reduce table to minimal number of columns, just source ID,
            # positions, and fluxes
            columns = ["id", "xcentroid", "ycentroid", "flux"]

            if sources:
                catalog = sources[columns]
                log.info(f"Found {len(catalog)} sources.")
            else:
                # if no sources were detected, return an empty table
                self.log.warning("No sources detected, returning empty catalog.")
                catalog = Table(
                    names=columns,
                    dtype=(int, np.float64, np.float64, np.float64),
                )

            if self.fit_psf and len(sources):
                # refine astrometry by fitting PSF models to the detected sources
                log.info("Constructing a gridded PSF model.")
                detector = input_model.meta.instrument["detector"].replace("WFI", "SCA")
                gridded_psf_model, _ = psf.create_gridded_psf_model(
                    path_prefix="tmp",
                    filt=filt,
                    detector=detector,
                    overwrite=True,
                    logging_level="ERROR",
                )

                log.info(
                    "Fitting a PSF model to sources for improved astrometric precision."
                )
                psf_photometry_table, photometry = psf.fit_psf_to_image_model(
                    image_model=input_model,
                    psf_model=gridded_psf_model,
                    x_init=catalog["xcentroid"],
                    y_init=catalog["ycentroid"],
                    exclude_out_of_bounds=True,
                )

            catalog_as_recarray = catalog.as_array()

            # create meta.source detection section in file
            # if save_catalogs is True, this will be updated with the
            # attribute 'tweakreg_catalog_name' to point to the location
            # of the catalog on disk. If save_catalogs is false, this section
            # will be updated to contain the catalog to pass to TweakReg

            # tweakreg_catalog_name will be saved to the final output file,
            # while tweakreg_catalog is intended to be deleted by TweakRegStep
            input_model.meta["source_detection"] = maker_utils.mk_source_detection()

            # if 'save_catalogs'= True, also save the output catalog to a file
            # (format specified by output_cat_filetype) and add an attribute
            # to the file that contains the path to this file
            if self.save_catalogs:
                cat_filename = input_model.meta.filename.replace(".asdf", "")
                cat_filename += f"_tweakreg_catalog.{self.output_cat_filetype}"
                log.info(f"Saving catalog to file: {cat_filename}.")

                if self.output_cat_filetype == "asdf":
                    tree = {"tweakreg_catalog": catalog_as_recarray}
                    ff = AsdfFile(tree)
                    ff.write_to(cat_filename)
                else:
                    catalog.write(cat_filename, format="ascii.ecsv", overwrite=True)

                input_model.meta.source_detection["tweakreg_catalog_name"] = (
                    cat_filename
                )
            else:
                if self.fit_psf:
                    # PSF photometry centroid results are stored in an astropy table
                    # in columns "x_fit" and "y_fit", which we'll rename for
                    # compatibility with tweakreg:
                    catalog = psf_photometry_table.copy()

                    # rename the PSF photometry table's columns to match the
                    # expectated columns in tweakreg:
                    for old_name, new_name in zip(
                        ["x_fit", "y_fit"], ["xcentroid", "ycentroid"]
                    ):
                        catalog.rename_column(old_name, new_name)

                    input_model.meta.source_detection["tweakreg_catalog"] = catalog
                else:
                    # only attach catalog to file if its being passed to the next step
                    # and save_catalogs is false, since it is not in the schema
                    input_model.meta.source_detection["tweakreg_catalog"] = (
                        catalog_as_recarray
                    )
            input_model.meta.cal_step["source_detection"] = "COMPLETE"

        # just pass input model to next step - catalog is stored in meta
        return input_model

    def _calc_2D_background(self):
        """Calculates a 2D background image.

        Calculates the background value for the input image in boxes specified by
        self.bkg_box_size. A mean, median, or mode estimator may be used (set
        by `bkg_estimator`). The pixels in each box will be sigma clipped,
        using a sigma specified by `bkg_sigma`."""

        filter_size = (
            self.bkg_filter_size,
            self.bkg_filter_size,
        )  # square size specified
        box_size = np.asarray(self.bkg_boxsize).astype(int)  # must be integer

        if self.bkg_estimator == "median":
            bkg_estimator = MedianBackground()
        elif self.bkg_estimator == "mean":
            bkg_estimator = MeanBackground()
        elif self.bkg_estimator == "mode":
            bkg_estimator = ModeEstimatorBackground()
        else:
            raise ValueError("bkg_estimator must be one of 'mean', 'median', or 'mode'")

        sigma_clip = SigmaClip(self.bkg_sigma)

        try:
            bkg_2D = Background2D(
                self.data,
                box_size,
                filter_size=filter_size,
                coverage_mask=self.coverage_mask,
                sigma_clip=sigma_clip,
                bkg_estimator=bkg_estimator,
            )
        except ValueError:
            # use the entire unmasked array
            log.info(
                "Background could not be estimated in meshes. "
                "Using the entire unmasked array for background "
                f"estimation: bkg_boxsize={self.data.shape}."
            )

            bkg_2D = Background2D(
                self.data,
                self.data.shape,
                filter_size=filter_size,
                coverage_mask=self.coverage_mask,
                sigma_clip=sigma_clip,
                bkg_estimator=bkg_estimator,
                exclude_percentile=100.0,
            )

        return bkg_2D
